from unittest import mock

import pytest

from orchestrator import app_settings
from orchestrator.graphql.autoregistration import register_domain_models
from test.unit_tests.fixtures.products.product_blocks.product_block_list_nested import (
    ProductBlockListNestedForTestInactiveGraphql,
)


@pytest.fixture(autouse=True)
def fastapi_app_graphql(
    fastapi_app,
    test_client,
    generic_product_type_1,
    generic_product_type_2,
    test_product_type_sub_list_union,
    test_product_type_sub_one,
    test_product_type_sub_two,
    product_type_1_subscriptions_factory,
    sub_one_subscription_1,
    sub_two_subscription_1,
    product_sub_list_union_subscription_1,
):
    from pydantic import BaseModel

    from orchestrator.graphql.schemas.subscription import MetadataDict

    class Metadata(BaseModel):
        some_metadata_prop: list[str]

    MetadataDict.update({"metadata": Metadata})

    actual_env = app_settings.ENVIRONMENT
    app_settings.ENVIRONMENT = "TESTING"
    fastapi_app.register_graphql()
    yield fastapi_app
    app_settings.ENVIRONMENT = actual_env


@pytest.fixture(scope="session", autouse=True)
def fix_graphql_model_registration():
    # This block caches the "ProductBlockListNestedForTestInactive" model to avoid re-instantiation in each test case.
    # This is necessary because this product block has a self referencing property, which strawberry can't handle correctly,
    # and lead to an error expecting the `ProductBlockListNestedForTestInactive` strawberry type to already exist.
    internal_graphql_models = {"ProductBlockListNestedForTestInactive": ProductBlockListNestedForTestInactiveGraphql}

    def patched_register_domain_models(*args, **kwargs):
        graphql_models = register_domain_models(*args, **kwargs)
        internal_graphql_models.update(graphql_models)
        return internal_graphql_models

    with mock.patch("orchestrator.graphql.schema.register_domain_models", side_effect=patched_register_domain_models):
        yield



from orchestrator.graphql.autoregistration import create_strawberry_enums
from test.unit_tests.fixtures.products.product_blocks.product_block_one import DummyEnum


def test_create_strawberry_enums(test_product_block_one):
    _, _, ProductBlockOneForTest = test_product_block_one
    assert create_strawberry_enums(ProductBlockOneForTest, {}) == {"enum_field": DummyEnum}


def test_create_strawberry_enums_optional(test_product_block_one):
    ProductBlockOneForTestInactive, _, _ = test_product_block_one
    assert create_strawberry_enums(ProductBlockOneForTestInactive, {}) == {"enum_field": DummyEnum}



from typing import Annotated

import structlog
from orchestrator.domain import SubscriptionModel
from orchestrator.forms import FormPage
from orchestrator.forms.validators import CustomerId, Divider, Label
from orchestrator.targets import Target
from orchestrator.types import SubscriptionLifecycle
from orchestrator.workflow import StepList, begin, step
from orchestrator.workflows.steps import store_process_subscription
from orchestrator.workflows.utils import create_workflow
from pydantic import AfterValidator, ConfigDict
from pydantic_forms.types import FormGenerator, State, UUIDstr

from products.product_blocks.example1 import AnnotatedInt, ExampleStrEnum1
from products.product_types.example1 import Example1Inactive, Example1Provisioning
from workflows.example1.shared.forms import (
    annotated_int_must_be_unique_validator,
    must_be_unused_to_change_mode_validator,
)
from workflows.shared import create_summary_form


def subscription_description(subscription: SubscriptionModel) -> str:
    """Generate subscription description.

    The suggested pattern is to implement a subscription service that generates a subscription specific
    description, in case that is not present the description will just be set to the product name.
    """
    return f"{subscription.product.name} subscription"


logger = structlog.get_logger(__name__)

validated_example_str_enum_1 = Annotated[ExampleStrEnum1, AfterValidator(must_be_unused_to_change_mode_validator)]
validated_annotated_int = Annotated[AnnotatedInt, AfterValidator(annotated_int_must_be_unique_validator)]


def initial_input_form_generator(product_name: str) -> FormGenerator:
    # TODO add additional fields to form if needed

    class CreateExample1Form(FormPage):
        model_config = ConfigDict(title=product_name)

        customer_id: CustomerId

        example1_settings: Label
        divider_1: Divider

        example_str_enum_1: validated_example_str_enum_1
        unmodifiable_str: str
        modifiable_boolean: bool
        annotated_int: validated_annotated_int | None = None
        always_optional_str: str | None = None

    user_input = yield CreateExample1Form
    user_input_dict = user_input.dict()

    summary_fields = [
        "example_str_enum_1",
        "unmodifiable_str",
        "modifiable_boolean",
        "annotated_int",
        "always_optional_str",
    ]
    yield from create_summary_form(user_input_dict, product_name, summary_fields)

    return user_input_dict


@step("Construct Subscription model")
def construct_example1_model(
    product: UUIDstr,
    customer_id: UUIDstr,
    example_str_enum_1: ExampleStrEnum1,
    unmodifiable_str: str,
    modifiable_boolean: bool,
    annotated_int: AnnotatedInt | None,
    always_optional_str: str | None,
) -> State:
    example1 = Example1Inactive.from_product_id(
        product_id=product,
        customer_id=customer_id,
        status=SubscriptionLifecycle.INITIAL,
    )
    example1.example1.example_str_enum_1 = example_str_enum_1
    example1.example1.unmodifiable_str = unmodifiable_str
    example1.example1.modifiable_boolean = modifiable_boolean
    example1.example1.annotated_int = annotated_int
    example1.example1.always_optional_str = always_optional_str

    example1 = Example1Provisioning.from_other_lifecycle(example1, SubscriptionLifecycle.PROVISIONING)
    example1.description = subscription_description(example1)

    return {
        "subscription": example1,
        "subscription_id": example1.subscription_id,  # necessary to be able to use older generic step functions
        "subscription_description": example1.description,
    }


additional_steps = begin


@create_workflow("Create example1", initial_input_form=initial_input_form_generator, additional_steps=additional_steps)
def create_example1() -> StepList:
    return (
        begin >> construct_example1_model >> store_process_subscription(Target.CREATE)
        # TODO add provision step(s)
    )



from test.unit_tests.metrics.conftest import EMPTY_METRICS


def test_engine_metrics_success(test_client) -> None:
    test_client.put("/api/settings/status", json={"global_lock": True})
    response = test_client.get("api/metrics")
    expected_metric_lines = [
        "# HELP wfo_engine_status Current workflow engine status.",
        "# TYPE wfo_engine_status gauge",
        'wfo_engine_status{wfo_engine_status="PAUSED"} 1.0',
        'wfo_engine_status{wfo_engine_status="PAUSING"} 0.0',
        'wfo_engine_status{wfo_engine_status="RUNNING"} 0.0',
    ]
    assert all(line in response.text for line in expected_metric_lines)

    test_client.put("/api/settings/status", json={"global_lock": False})
    response = test_client.get("/api/metrics")
    assert response.text == EMPTY_METRICS



import pytest

from orchestrator.db import ProductTable, db
from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY
from orchestrator.domain.base import SubscriptionModel
from orchestrator.types import SubscriptionLifecycle


@pytest.fixture
def test_union_type_product(test_product_block_one, test_product_sub_block_one):
    ProductBlockOneForTestInactive, ProductBlockOneForTestProvisioning, ProductBlockOneForTest = test_product_block_one
    SubBlockOneForTestInactive, SubBlockOneForTestProvisioning, SubBlockOneForTest = test_product_sub_block_one

    class UnionProductInactive(SubscriptionModel, is_base=True):
        test_block: ProductBlockOneForTestInactive | None
        union_block: ProductBlockOneForTestInactive | SubBlockOneForTestInactive | None

    class UnionProductProvisioning(UnionProductInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
        test_block: ProductBlockOneForTestProvisioning
        union_block: ProductBlockOneForTestProvisioning | SubBlockOneForTestProvisioning

    class UnionProduct(UnionProductProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
        test_block: ProductBlockOneForTest
        union_block: ProductBlockOneForTest | SubBlockOneForTest

    SUBSCRIPTION_MODEL_REGISTRY["UnionProduct"] = UnionProduct
    yield UnionProductInactive, UnionProductProvisioning, UnionProduct
    del SUBSCRIPTION_MODEL_REGISTRY["UnionProduct"]


@pytest.fixture
def test_union_product(test_product_block_one_db):
    product = ProductTable(
        name="UnionProduct", description="Test Union Product", product_type="Test", tag="Union", status="active"
    )

    product_block, product_sub_block = test_product_block_one_db
    product.product_blocks = [product_block, product_sub_block]
    db.session.add(product)
    db.session.commit()
    return product.product_id



from uuid import uuid4

import pytest
from pydantic import ValidationError

from orchestrator.db import db
from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SubscriptionLifecycle
from test.unit_tests.fixtures.products.product_blocks.product_block_one import DummyEnum


def blocks_sorted(blocks):
    def sort_key(block: ProductBlockModel) -> tuple:
        return (block.owner_subscription_id, block.subscription_instance_id)

    return sorted(blocks, key=sort_key, reverse=True)


def test_product_model_with_list_union_type_directly_below(
    test_product_list_union,
    test_product_type_list_union,
    test_product_sub_block_two,
    sub_two_subscription_1,
):
    ProductListUnionInactive, _, ProductListUnion = test_product_type_list_union
    _, _, SubBlockTwoForTest = test_product_sub_block_two

    list_union_subscription_inactive = ProductListUnionInactive.from_product_id(
        product_id=test_product_list_union, customer_id=str(uuid4())
    )

    with pytest.raises(ValidationError):
        ProductListUnion.from_other_lifecycle(list_union_subscription_inactive, SubscriptionLifecycle.ACTIVE)

    new_sub_block_1 = SubBlockTwoForTest.new(
        subscription_id=list_union_subscription_inactive.subscription_id, int_field_2=1
    )
    new_sub_block_2 = SubBlockTwoForTest.new(
        subscription_id=list_union_subscription_inactive.subscription_id, int_field_2=2
    )
    list_union_subscription_inactive.list_union_blocks = [new_sub_block_1, new_sub_block_2]
    list_union_subscription_inactive.save()

    assert (
        list_union_subscription_inactive.diff_product_in_database(list_union_subscription_inactive.product.product_id)
        == {}
    )
    list_union_subscription = ProductListUnion.from_other_lifecycle(
        list_union_subscription_inactive, SubscriptionLifecycle.ACTIVE
    )
    list_union_subscription.save()

    list_union_sub_from_database = ProductListUnion.from_subscription(list_union_subscription.subscription_id)
    assert type(list_union_sub_from_database) is type(list_union_subscription)

    sorted_db_list = blocks_sorted(list_union_sub_from_database.list_union_blocks)
    sorted_sub_list = blocks_sorted(list_union_subscription.list_union_blocks)
    assert sorted_db_list == sorted_sub_list

    list_union_subscription.list_union_blocks = [sub_two_subscription_1.test_block]

    with pytest.raises(ValueError) as exc:
        list_union_subscription.save()
        assert (
            str(exc)
            == "Attempting to save a Foreign `Subscription Instance` directly below a subscription. This is not allowed."
        )


def test_product_model_with_list_union_type_directly_below_with_relation_overlap(
    test_product_list_union_overlap,
    test_product_type_list_union_overlap,
    test_product_sub_block_one,
    test_product_block_one,
    sub_one_subscription_1,
):
    ProductListUnionInactive, _, ProductListUnion = test_product_type_list_union_overlap
    SubBlockOneForTestInactive, _, _ = test_product_sub_block_one
    ProductBlockOneForTestInactive, _, _ = test_product_block_one

    list_union_subscription_inactive = ProductListUnionInactive.from_product_id(
        product_id=test_product_list_union_overlap, customer_id=str(uuid4())
    )

    list_union_subscription_inactive.test_block = ProductBlockOneForTestInactive.new(
        subscription_id=list_union_subscription_inactive.subscription_id,
        int_field=3,
        str_field="",
        list_field=[1],
        enum_field=DummyEnum.FOO,
        sub_block=SubBlockOneForTestInactive.new(
            subscription_id=list_union_subscription_inactive.subscription_id, int_field=3, str_field="2"
        ),
        sub_block_2=SubBlockOneForTestInactive.new(
            subscription_id=list_union_subscription_inactive.subscription_id, int_field=3, str_field="2"
        ),
    )

    with pytest.raises(ValidationError):
        ProductListUnion.from_other_lifecycle(list_union_subscription_inactive, SubscriptionLifecycle.ACTIVE)

    new_sub_block_1 = SubBlockOneForTestInactive.new(
        subscription_id=list_union_subscription_inactive.subscription_id, int_field=11, str_field="111"
    )
    new_sub_block_2 = SubBlockOneForTestInactive.new(
        subscription_id=list_union_subscription_inactive.subscription_id, int_field=12, str_field="121"
    )
    list_union_subscription_inactive.list_union_blocks = [new_sub_block_1, new_sub_block_2]
    list_union_subscription_inactive.save()

    assert (
        list_union_subscription_inactive.diff_product_in_database(list_union_subscription_inactive.product.product_id)
        == {}
    )
    list_union_subscription = ProductListUnion.from_other_lifecycle(
        list_union_subscription_inactive, SubscriptionLifecycle.ACTIVE
    )
    list_union_subscription.save()

    list_union_sub_from_database = ProductListUnion.from_subscription(list_union_subscription.subscription_id)
    assert type(list_union_sub_from_database) is type(list_union_subscription)
    assert list_union_sub_from_database.test_block == list_union_subscription.test_block

    sorted_db_list_len = len(list_union_sub_from_database.list_union_blocks)
    sorted_sub_list_len = len(list_union_subscription.list_union_blocks)
    assert sorted_db_list_len != sorted_sub_list_len
    assert sorted_db_list_len == 5  # 3 were made with test_block, which also get included.

    list_union_subscription.list_union_blocks = [sub_one_subscription_1.test_block]

    with pytest.raises(ValueError) as exc:
        list_union_subscription.save()
        assert (
            str(exc)
            == "Attempting to save a Foreign `Subscription Instance` directly below a subscription. This is not allowed."
        )


def test_list_union_product_block_as_sub(
    test_product_sub_list_union,
    test_product_type_sub_list_union,
    test_product_block_with_list_union,
    test_product_type_sub_one,
    sub_one_subscription_1,
    sub_two_subscription_1,
):
    ProductSubListUnionInactive, _, ProductSubListUnion = test_product_type_sub_list_union
    ProductListUnionBlockForTestInactive, _, _ = test_product_block_with_list_union
    _, _, ProductSubOne = test_product_type_sub_one

    list_union_subscription_inactive = ProductSubListUnionInactive.from_product_id(
        product_id=test_product_sub_list_union, customer_id=str(uuid4())
    )
    list_union_subscription_inactive.test_block = ProductListUnionBlockForTestInactive.new(
        subscription_id=list_union_subscription_inactive.subscription_id
    )
    list_union_subscription_inactive.save()

    list_union_subscription_inactive.test_block.int_field = 1
    list_union_subscription_inactive.test_block.str_field = "blah"
    list_union_subscription_inactive.test_block.list_union_blocks = [
        sub_one_subscription_1.test_block,
        sub_two_subscription_1.test_block,
    ]

    list_union_subscription_inactive.test_block.list_field = [2]

    list_union_subscription = ProductSubListUnion.from_other_lifecycle(
        list_union_subscription_inactive, status=SubscriptionLifecycle.ACTIVE
    )
    list_union_subscription.save()

    # This needs to happen in the test due to the fact it is using cached objects.
    db.session.commit()
    assert list_union_subscription.diff_product_in_database(test_product_sub_list_union) == {}

    list_union_sub_from_database = ProductSubListUnion.from_subscription(list_union_subscription.subscription_id)

    assert type(list_union_sub_from_database) is type(list_union_subscription)
    assert list_union_sub_from_database.test_block.int_field == list_union_subscription.test_block.int_field
    assert list_union_sub_from_database.test_block.str_field == list_union_subscription.test_block.str_field

    sorted_db_list = blocks_sorted(list_union_sub_from_database.test_block.list_union_blocks)
    sorted_sub_list = blocks_sorted(list_union_subscription_inactive.test_block.list_union_blocks)
    assert sorted_db_list == sorted_sub_list

    # Do not allow subscriptions that are in use by other subscriptions make an unsafe transition.
    with pytest.raises(ValueError):
        ProductSubOne.from_other_lifecycle(sub_one_subscription_1, SubscriptionLifecycle.TERMINATED)



from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SubscriptionLifecycle
from pydantic import computed_field


class Example4SubBlockInactive(ProductBlockModel, product_block_name="Example4Sub"):
    str_val: str | None = None


class Example4SubBlockProvisioning(Example4SubBlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
    str_val: str | None = None

    @computed_field
    @property
    def title(self) -> str:
        # TODO: format correct title string
        return f"{self.name}"


class Example4SubBlock(Example4SubBlockProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    str_val: str | None = None



import json
from http import HTTPStatus
from uuid import UUID

import pytest
import strawberry

from orchestrator.graphql.schemas.customer import CustomerType
from orchestrator.graphql.utils.override_class import override_class
from test.unit_tests.config import GRAPHQL_ENDPOINT, GRAPHQL_HEADERS


@pytest.fixture
def override_class_app_graphql(
    fastapi_app_graphql,
    test_client,
    sub_one_subscription_1,
):
    def customer_id_override(self) -> str:
        return "overriden"

    customer_id_override_field = strawberry.field(resolver=customer_id_override, description="Returns customer_id")
    customer_id_override_field.name = "customer_id"

    def new(self) -> UUID:
        return self.customer_id

    new_field = strawberry.field(resolver=new, description="Returns new_field")
    new_field.name = "new_field"

    override_class(CustomerType, [customer_id_override_field, new_field])

    fastapi_app_graphql.register_graphql()
    yield fastapi_app_graphql

    # reset the overriden fields.

    @strawberry.field(description="Returns customer_id")  # type: ignore
    def customer_id(self) -> str:
        return self.customer_id

    customer_id.name = "customer_id"

    override_class(CustomerType, [customer_id])
    CustomerType.__strawberry_definition__.fields = [
        field for field in CustomerType.__strawberry_definition__.fields if field.name != "new_field"
    ]
    fastapi_app_graphql.register_graphql()


def get_customers_query(
    first: int = 1,
    after: int = 0,
    filter_by: list[str] | None = None,
    sort_by: list[dict[str, str]] | None = None,
    override_fields: bool = False,
) -> bytes:
    query = """
query CustomerQuery($first: Int!, $after: Int!, $filterBy: [GraphqlFilter!], $sortBy: [GraphqlSort!]) {{
  customers(first: $first, after: $after, filterBy: $filterBy, sortBy: $sortBy) {{
    page {{
      customerId
      fullname
      shortcode
        {}
    }}
  }}
}}
    """.format(
        "newField" if override_fields else "",
    )
    return json.dumps(
        {
            "operationName": "CustomerQuery",
            "query": query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
            },
        }
    ).encode("utf-8")


def get_subscriptions_customer_query(
    first: int = 1,
    after: int = 0,
    filter_by: list[str] | None = None,
    sort_by: list[dict[str, str]] | None = None,
    override_fields: bool = False,
) -> bytes:
    query = """
query SubscriptionQuery($first: Int!, $after: Int!, $filterBy: [GraphqlFilter!], $sortBy: [GraphqlSort!]) {{
  subscriptions(first: $first, after: $after, filterBy: $filterBy, sortBy: $sortBy) {{
    page {{
      customer {{
        customerId
        fullname
        shortcode
        {}
      }}
    }}
  }}
}}
    """.format(
        "newField" if override_fields else "",
    )
    return json.dumps(
        {
            "operationName": "SubscriptionQuery",
            "query": query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
            },
        }
    ).encode("utf-8")


def get_subscriptions_detail_customer_query(
    first: int = 1,
    after: int = 0,
    filter_by: list[str] | None = None,
    sort_by: list[dict[str, str]] | None = None,
    override_fields: bool = False,
) -> bytes:
    query = """
query SubscriptionQuery($first: Int!, $after: Int!, $filterBy: [GraphqlFilter!], $sortBy: [GraphqlSort!]) {{
  subscriptions(first: $first, after: $after, filterBy: $filterBy, sortBy: $sortBy) {{
    page {{
      ... on ProductSubOneSubscription {{
        customer {{
          customerId
          fullname
          shortcode
          {}
        }}
      }}
    }}
  }}
}}
    """.format(
        "newField" if override_fields else "",
    )
    return json.dumps(
        {
            "operationName": "SubscriptionQuery",
            "query": query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
            },
        }
    ).encode("utf-8")


def test_customers(fastapi_app_graphql, test_client):
    # when
    response = test_client.post(GRAPHQL_ENDPOINT, content=get_customers_query(), headers=GRAPHQL_HEADERS)

    # then
    assert HTTPStatus.OK == response.status_code
    result = response.json()
    customer_data = result["data"]["customers"]["page"][0]
    assert customer_data == {
        "customerId": "59289a57-70fb-4ff5-9c93-10fe67b12434",
        "fullname": "Default::Orchestrator-Core Customer",
        "shortcode": "default-cust",
    }


def test_customers_overriden(override_class_app_graphql, test_client):
    # when
    response = test_client.post(
        GRAPHQL_ENDPOINT, content=get_customers_query(override_fields=True), headers=GRAPHQL_HEADERS
    )

    # then
    assert HTTPStatus.OK == response.status_code
    result = response.json()
    customer_data = result["data"]["customers"]["page"][0]
    assert customer_data == {
        "customerId": "overriden",
        "fullname": "Default::Orchestrator-Core Customer",
        "shortcode": "default-cust",
        "newField": "59289a57-70fb-4ff5-9c93-10fe67b12434",
    }


def test_subscription_customer(fastapi_app_graphql, test_client, sub_one_subscription_1):
    # when
    response = test_client.post(
        GRAPHQL_ENDPOINT,
        content=get_subscriptions_customer_query(
            filter_by=[{"field": "subscriptionId", "value": str(sub_one_subscription_1.subscription_id)}]
        ),
        headers=GRAPHQL_HEADERS,
    )

    # then
    assert HTTPStatus.OK == response.status_code
    result = response.json()
    customer_data = result["data"]["subscriptions"]["page"][0]
    assert customer_data == {
        "customer": {
            "customerId": "59289a57-70fb-4ff5-9c93-10fe67b12434",
            "fullname": "Default::Orchestrator-Core Customer",
            "shortcode": "default-cust",
        }
    }


def test_subscription_customer_overriden(override_class_app_graphql, test_client, sub_one_subscription_1):
    # when
    response = test_client.post(
        GRAPHQL_ENDPOINT,
        content=get_subscriptions_customer_query(
            filter_by=[{"field": "subscriptionId", "value": str(sub_one_subscription_1.subscription_id)}],
            override_fields=True,
        ),
        headers=GRAPHQL_HEADERS,
    )

    # then
    assert HTTPStatus.OK == response.status_code
    result = response.json()
    customer_data = result["data"]["subscriptions"]["page"][0]

    assert customer_data == {
        "customer": {
            "customerId": "overriden",
            "fullname": "Default::Orchestrator-Core Customer",
            "shortcode": "default-cust",
            "newField": "59289a57-70fb-4ff5-9c93-10fe67b12434",
        }
    }


def test_subscription_detail_customer(fastapi_app_graphql, test_client, sub_one_subscription_1):
    # when
    response = test_client.post(
        GRAPHQL_ENDPOINT,
        content=get_subscriptions_detail_customer_query(
            filter_by=[{"field": "subscriptionId", "value": str(sub_one_subscription_1.subscription_id)}]
        ),
        headers=GRAPHQL_HEADERS,
    )

    # then
    assert HTTPStatus.OK == response.status_code
    result = response.json()
    customer_data = result["data"]["subscriptions"]["page"][0]
    assert customer_data == {
        "customer": {
            "customerId": "59289a57-70fb-4ff5-9c93-10fe67b12434",
            "fullname": "Default::Orchestrator-Core Customer",
            "shortcode": "default-cust",
        }
    }


def test_subscription_detail_customer_overriden(override_class_app_graphql, test_client, sub_one_subscription_1):
    # when
    response = test_client.post(
        GRAPHQL_ENDPOINT,
        content=get_subscriptions_detail_customer_query(
            filter_by=[{"field": "subscriptionId", "value": str(sub_one_subscription_1.subscription_id)}],
            override_fields=True,
        ),
        headers=GRAPHQL_HEADERS,
    )

    # then
    assert HTTPStatus.OK == response.status_code
    result = response.json()
    customer_data = result["data"]["subscriptions"]["page"][0]

    assert customer_data == {
        "customer": {
            "customerId": "overriden",
            "fullname": "Default::Orchestrator-Core Customer",
            "shortcode": "default-cust",
            "newField": "59289a57-70fb-4ff5-9c93-10fe67b12434",
        }
    }



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from orchestrator.domain.base import SubscriptionModel
from orchestrator.types import SubscriptionLifecycle
from test_orchestrator.product_blocks.test_product_blocks import (
    TestProductBlock,
    TestProductBlockInactive,
    TestProductBlockProvisioning,
)


class TestProductInactive(SubscriptionModel, is_base=True):
    testproduct: TestProductBlockInactive


class TestProductProvisioning(TestProductInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
    testproduct: TestProductBlockProvisioning


class TestProduct(TestProductProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    testproduct: TestProductBlock



import pytest

from test.unit_tests.workflows import assert_complete, extract_state, run_workflow


@pytest.mark.workflow
def test_happy_flow(responses, example1_subscription):
    # when

    result, _, _ = run_workflow("validate_example1", {"subscription_id": example1_subscription})

    # then

    assert_complete(result)
    state = extract_state(result)
    assert state["check_core_db"] is True


@pytest.mark.workflow
def test_validate_example_in_some_oss(responses, example1_subscription):
    # given

    # TODO: set test conditions or fixture so that "Validate that the example1 subscription is correctly administered in some external system" triggers

    # when

    with pytest.raises(AssertionError) as error:
        result, _, _ = run_workflow("validate_example1", [{"subscription_id": example1_subscription}, {}])

    # then

    assert (
        error.value.errors[0]["msg"]
        == "Validate that the example1 subscription is correctly administered in some external system"
    )



# Copyright 2022 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from http import HTTPStatus

import pytest

from orchestrator.db.filters.process import process_filter_fields
from orchestrator.db.filters.product import product_filter_fields
from orchestrator.db.filters.product_block import product_block_filter_fields
from orchestrator.db.filters.resource_type import resource_type_filter_fields
from orchestrator.db.filters.subscription import subscription_filter_fields
from orchestrator.db.filters.workflow import workflow_filter_fields
from orchestrator.db.sorting.process import process_sort_fields
from orchestrator.db.sorting.product import product_sort_fields
from orchestrator.db.sorting.product_block import product_block_sort_fields
from orchestrator.db.sorting.resource_type import resource_type_sort_fields
from orchestrator.db.sorting.subscription import subscription_sort_fields
from orchestrator.db.sorting.workflow import workflow_sort_fields


def get_page_info_sort_and_filter_fields(type_name) -> bytes:
    query = f"""
query PageInfoSortAndFilterQuery {{
  {type_name} {{
    pageInfo {{
      sortFields
      filterFields
    }}
  }}
}}
"""
    return json.dumps(
        {
            "operationName": "PageInfoSortAndFilterQuery",
            "query": query,
        }
    ).encode("utf-8")


@pytest.mark.parametrize(
    "type_name, sort_fields, filter_fields",
    [
        ("processes", process_sort_fields(), process_filter_fields()),
        ("subscriptions", subscription_sort_fields(), subscription_filter_fields()),
        ("products", product_sort_fields(), product_filter_fields()),
        ("productBlocks", product_block_sort_fields(), product_block_filter_fields()),
        ("workflows", workflow_sort_fields(), workflow_filter_fields()),
        ("resourceTypes", resource_type_sort_fields(), resource_type_filter_fields()),
    ],
)
def test_process_sort_and_filter_fields_in_page_info(test_client, type_name, sort_fields, filter_fields):
    data = get_page_info_sort_and_filter_fields(type_name)
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    processes_data = result["data"][type_name]
    pageinfo = processes_data["pageInfo"]

    assert pageinfo == {
        "sortFields": sort_fields,
        "filterFields": filter_fields,
    }



from collections.abc import Generator
from typing import List, TypeAlias, cast

from orchestrator.domain.base import ProductBlockModel
from orchestrator.forms import FormPage
from orchestrator.forms.validators import MigrationSummary, migration_summary
from pydantic import ConfigDict


def summary_form(product_name: str, summary_data: dict) -> Generator:
    ProductSummary: TypeAlias = cast("type[MigrationSummary]", migration_summary(summary_data))

    class SummaryForm(FormPage):
        model_config = ConfigDict(title=f"{product_name} summary")

        product_summary: ProductSummary

    yield SummaryForm


def create_summary_form(user_input: dict, product_name: str, fields: List[str]) -> Generator:
    columns = [[str(user_input[nm]) for nm in fields]]
    yield from summary_form(product_name, {"labels": fields, "columns": columns})


def modify_summary_form(user_input: dict, block: ProductBlockModel, fields: List[str]) -> Generator:
    before = [str(getattr(block, nm)) for nm in fields]  # type: ignore[attr-defined]
    after = [str(user_input[nm]) for nm in fields]
    yield from summary_form(
        block.subscription.product.name,
        {
            "labels": fields,
            "headers": ["Before", "After"],
            "columns": [before, after],
        },
    )



# Copyright 2022 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from http import HTTPStatus

import pytest

from test.unit_tests.conftest import do_refresh_subscriptions_search_view


def assert_result_ids_against_expected_ids(result_ids, expected_ids):
    assert sorted(result_ids) == sorted([str(id) for id in expected_ids])


def build_subscriptions_relation_query(body: str) -> str:
    return f"""
query SubscriptionQuery(
    $first: Int!,
    $after: Int!,
    $sortBy: [GraphqlSort!],
    $filterBy: [GraphqlFilter!],
    $query: String,
    $dependsOnFilter: SubscriptionRelationFilter,
    $dependsOnSubscriptionsFilter: [GraphqlFilter!],
    $inUseByFilter: SubscriptionRelationFilter,
    $inUseBySubscriptionsFilter: [GraphqlFilter!],
) {{
    subscriptions(first: $first, after: $after, sortBy: $sortBy, filterBy: $filterBy, query: $query)
    {body}
}}
"""


def get_subscriptions_query_with_processes(
    first: int = 10,
    after: int = 0,
    filter_by: list[dict[str, str]] | None = None,
    sort_by: list[dict[str, str]] | None = None,
    query_string: str | None = None,
) -> bytes:
    query = """
query SubscriptionQuery(
    $first: Int!,
    $after: Int!,
    $sortBy: [GraphqlSort!],
    $filterBy: [GraphqlFilter!],
    $query: String,
) {
  subscriptions(first: $first, after: $after, sortBy: $sortBy, filterBy: $filterBy, query: $query) {
    page {
      description
      subscriptionId
      status
      insync
      note
      startDate
      endDate
      productBlockInstances {
        ownerSubscriptionId
        inUseByRelations
      }
      processes(sortBy: [{field: "startedAt", order: ASC}]) {
        page {
          processId
          isTask
          lastStep
          lastStatus
          assignee
          failedReason
          traceback
          workflowName
          createdBy
          startedAt
          lastModifiedAt
          product {
            productId
            name
            description
            productType
            status
            tag
            createdAt
            endDate
          }
        }
      }
    }
    pageInfo {
      startCursor
      totalItems
      hasPreviousPage
      endCursor
      hasNextPage
    }
  }
}
    """
    return json.dumps(
        {
            "operationName": "SubscriptionQuery",
            "query": query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
                "query": query_string,
            },
        }
    ).encode("utf-8")


def get_subscriptions_query_with_relations(
    first: int = 10,
    after: int = 0,
    filter_by: list[dict[str, str]] | None = None,
    sort_by: list[dict[str, str]] | None = None,
    query_string: str | None = None,
    depends_on_filter: dict[str, str] | None = None,
    depends_on_subscription_filter: dict[str, str] | None = None,
    in_use_by_filter: dict[str, str] | None = None,
    in_use_by_subscription_filter: dict[str, str] | None = None,
) -> bytes:
    query = build_subscriptions_relation_query(
        """{
    page {
      description
      subscriptionId
      status
      insync
      note
      startDate
      endDate
      dependsOnSubscriptions(dependsOnFilter: $dependsOnFilter, first: 20, filterBy: $dependsOnSubscriptionsFilter) {
        page {
          description
          subscriptionId
          status
          insync
          note
          startDate
          endDate
        }
      }
      inUseBySubscriptions(inUseByFilter: $inUseByFilter, first: 20, filterBy: $inUseBySubscriptionsFilter) {
        page {
          description
          subscriptionId
          status
          insync
          note
          startDate
          endDate
        }
      }
    }
    pageInfo {
      startCursor
      totalItems
      hasPreviousPage
      endCursor
      hasNextPage
    }
  }
    """
    )
    return json.dumps(
        {
            "operationName": "SubscriptionQuery",
            "query": query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
                "query": query_string,
                "dependsOnFilter": depends_on_filter,
                "dependsOnSubscriptionsFilter": depends_on_subscription_filter,
                "inUseByFilter": in_use_by_filter,
                "inUseBySubscriptionsFilter": in_use_by_subscription_filter,
            },
        }
    ).encode("utf-8")


@pytest.mark.parametrize(
    "query_args",
    [
        lambda sid: {"filter_by": [{"field": "subscriptionId", "value": sid}]},
        lambda sid: {"query_string": sid},
    ],
)
def test_single_subscription_with_processes(
    fastapi_app_graphql,
    test_client,
    product_type_1_subscriptions_factory,
    mocked_processes,
    mocked_processes_resumeall,  # noqa: F811
    generic_subscription_2,  # noqa: F811
    generic_subscription_1,
    query_args,
):
    # when

    product_type_1_subscriptions_factory(30)
    subscription_id = generic_subscription_1

    do_refresh_subscriptions_search_view()

    data = get_subscriptions_query_with_processes(**query_args(subscription_id))
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert len(subscriptions) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": 1,
    }
    assert subscriptions[0]["subscriptionId"] == subscription_id
    assert subscriptions[0]["processes"]["page"][0]["processId"] == str(mocked_processes[0])


def test_single_subscription_with_depends_on_subscriptions(
    fastapi_app_graphql,
    test_client,
    product_type_1_subscriptions_factory,
    sub_one_subscription_1,
    sub_two_subscription_1,
    product_sub_list_union_subscription_1,
):
    # when

    product_type_1_subscriptions_factory(30)

    do_refresh_subscriptions_search_view()

    subscription_id = str(product_sub_list_union_subscription_1)
    data = get_subscriptions_query_with_relations(query_string=subscription_id)
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    expected_depends_on_ids = {
        str(subscription.subscription_id) for subscription in [sub_one_subscription_1, sub_two_subscription_1]
    }
    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert len(subscriptions) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": 1,
    }
    assert subscriptions[0]["subscriptionId"] == subscription_id
    depends_on_ids = subscriptions[0]["dependsOnSubscriptions"]["page"]
    result_depends_on_ids = {subscription["subscriptionId"] for subscription in depends_on_ids}
    assert result_depends_on_ids == expected_depends_on_ids
    assert len(subscriptions[0]["inUseBySubscriptions"]["page"]) == 0


def test_single_subscription_with_in_use_by_subscriptions(
    fastapi_app_graphql,
    test_client,
    product_type_1_subscriptions_factory,
    sub_one_subscription_1,
    product_sub_list_union_subscription_1,
):
    # when

    product_type_1_subscriptions_factory(30)

    subscription_id = str(sub_one_subscription_1.subscription_id)
    subscription_query = get_subscriptions_query_with_relations(
        filter_by=[{"field": "subscriptionId", "value": subscription_id}]
    )

    response = test_client.post(
        "/api/graphql", content=subscription_query, headers={"Content-Type": "application/json"}
    )

    expected_in_use_by_ids = [str(product_sub_list_union_subscription_1)]

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert len(subscriptions) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": 1,
    }
    assert subscriptions[0]["subscriptionId"] == subscription_id
    assert len(subscriptions[0]["dependsOnSubscriptions"]["page"]) == 0
    result_in_use_by_ids = [
        subscription["subscriptionId"] for subscription in subscriptions[0]["inUseBySubscriptions"]["page"]
    ]
    assert result_in_use_by_ids == expected_in_use_by_ids


in_use_by_recurse_all_ids = [
    "subscription_20",
    "subscription_21",
    "subscription_22",
    "subscription_30",
    "subscription_31",
    "subscription_32",
    "subscription_33",
    "subscription_34",
    "subscription_40",
    "subscription_41",
    "subscription_42",
    "subscription_43",
    "subscription_44",
    "subscription_45",
]
in_use_by_recurse_all_with_filter_by_ids = [
    "subscription_22",
    "subscription_32",
]
in_use_by_recurse_status_active_ids = [
    "subscription_20",
    "subscription_22",
    "subscription_30",
    "subscription_32",
    "subscription_34",
    "subscription_40",
    "subscription_43",
    "subscription_45",
]
in_use_by_recurse_depth_limit_ids = [
    "subscription_20",
    "subscription_21",
    "subscription_22",
    "subscription_30",
    "subscription_31",
    "subscription_32",
    "subscription_33",
    "subscription_34",
]
in_use_by_recurse_only_product_type_test_ids = [
    "subscription_20",
    "subscription_21",
    "subscription_22",
    "subscription_30",
    "subscription_31",
    "subscription_32",
    "subscription_33",
    "subscription_40",
    "subscription_41",
    "subscription_42",
    "subscription_44",
]


@pytest.mark.parametrize(
    "in_use_by_filter,in_use_by_subscription_filter,subscription_ids,subscription_statuses",
    [
        (
            {"recurseProductTypes": ["Test", "ProductTypeOne", "ProductTypeTwo"]},
            None,
            in_use_by_recurse_all_ids,
            {"ACTIVE", "TERMINATED"},
        ),
        (
            {"recurseProductTypes": ["Test", "ProductTypeOne", "ProductTypeTwo"]},
            [{"field": "product", "value": "TestProductListNestedTypeOne"}],
            in_use_by_recurse_all_with_filter_by_ids,
            {"ACTIVE"},
        ),
        (
            {"statuses": ["active"], "recurseProductTypes": ["Test", "ProductTypeOne", "ProductTypeTwo"]},
            None,
            in_use_by_recurse_status_active_ids,
            {"ACTIVE"},
        ),
        (
            {"recurseProductTypes": ["Test", "ProductTypeOne", "ProductTypeTwo"], "recurseDepthLimit": 1},
            None,
            in_use_by_recurse_depth_limit_ids,
            {"ACTIVE", "TERMINATED"},
        ),
        (
            {"recurseProductTypes": ["Test"]},
            None,
            in_use_by_recurse_only_product_type_test_ids,
            {"ACTIVE", "TERMINATED"},
        ),
    ],
    ids=[
        "recurse_all",
        "recurse_all_with_filter_by",
        "recurse_status_active",
        "recurse_depth_limit",
        "recurse_only_product_type_test",
    ],
)
def test_single_subscription_with_in_use_by_subscriptions_recurse(
    fastapi_app_graphql,
    test_client,
    factory_subscription_with_nestings_in_use_by,
    in_use_by_filter,
    in_use_by_subscription_filter,
    subscription_ids,
    subscription_statuses,
):
    # when

    all_ids = factory_subscription_with_nestings_in_use_by

    subscription_id = str(all_ids["subscription_10"])
    subscription_query = get_subscriptions_query_with_relations(
        filter_by=[{"field": "subscriptionId", "value": subscription_id}],
        in_use_by_filter=in_use_by_filter,
        in_use_by_subscription_filter=in_use_by_subscription_filter,
    )

    response = test_client.post(
        "/api/graphql", content=subscription_query, headers={"Content-Type": "application/json"}
    )

    expected_relation_ids = [all_ids[sub_id] for sub_id in subscription_ids]

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]

    assert "errors" not in result
    assert subscriptions[0]["subscriptionId"] == subscription_id
    result_related_ids = [
        subscription["subscriptionId"] for subscription in subscriptions[0]["inUseBySubscriptions"]["page"]
    ]
    assert_result_ids_against_expected_ids(result_related_ids, expected_relation_ids)
    result_related_statuses = {
        subscription["status"] for subscription in subscriptions[0]["inUseBySubscriptions"]["page"]
    }
    assert result_related_statuses == subscription_statuses


depends_on_recurse_all_ids = [
    "subscription_10",
    "subscription_11",
    "subscription_12",
    "subscription_13",
    "subscription_14",
    "subscription_15",
    "subscription_20",
    "subscription_21",
    "subscription_22",
    "subscription_23",
    "subscription_24",
    "subscription_30",
    "subscription_31",
    "subscription_32",
]
depends_on_recurse_all_with_filter_by_ids = [
    "subscription_22",
    "subscription_32",
]
depends_on_recurse_status_active_ids = [
    "subscription_10",
    "subscription_13",
    "subscription_15",
    "subscription_20",
    "subscription_22",
    "subscription_24",
    "subscription_30",
    "subscription_32",
]
depends_on_recurse_depth_limit_ids = [
    "subscription_20",
    "subscription_21",
    "subscription_22",
    "subscription_23",
    "subscription_24",
    "subscription_30",
    "subscription_31",
    "subscription_32",
]
depends_on_recurse_only_product_type_test_ids = [
    "subscription_10",
    "subscription_11",
    "subscription_12",
    "subscription_14",
    "subscription_20",
    "subscription_21",
    "subscription_22",
    "subscription_23",
    "subscription_30",
    "subscription_31",
    "subscription_32",
]


@pytest.mark.parametrize(
    "depends_on_filter,depends_on_subscription_filter,subscription_ids,subscription_statuses",
    [
        (
            {"recurseProductTypes": ["Test", "ProductTypeOne", "ProductTypeTwo"]},
            None,
            depends_on_recurse_all_ids,
            {"ACTIVE", "TERMINATED"},
        ),
        (
            {"recurseProductTypes": ["Test", "ProductTypeOne", "ProductTypeTwo"]},
            [{"field": "product", "value": "TestProductListNestedTypeOne"}],
            depends_on_recurse_all_with_filter_by_ids,
            {"ACTIVE"},
        ),
        (
            {"statuses": ["active"], "recurseProductTypes": ["Test", "ProductTypeOne", "ProductTypeTwo"]},
            None,
            depends_on_recurse_status_active_ids,
            {"ACTIVE"},
        ),
        (
            {"recurseProductTypes": ["Test", "ProductTypeOne", "ProductTypeTwo"], "recurseDepthLimit": 1},
            None,
            depends_on_recurse_depth_limit_ids,
            {"ACTIVE", "TERMINATED"},
        ),
        (
            {"recurseProductTypes": ["Test"]},
            None,
            depends_on_recurse_only_product_type_test_ids,
            {"ACTIVE", "TERMINATED"},
        ),
    ],
    ids=[
        "recurse_all",
        "recurse_all_with_filter_by",
        "recurse_status_active",
        "recurse_depth_limit",
        "recurse_only_product_type_test",
    ],
)
def test_single_subscription_with_depends_on_subscriptions_recurse(
    fastapi_app_graphql,
    test_client,
    factory_subscription_with_nestings_depends_on,
    depends_on_filter,
    depends_on_subscription_filter,
    subscription_ids,
    subscription_statuses,
):
    # when

    all_ids = factory_subscription_with_nestings_depends_on

    subscription_id = str(all_ids["subscription_40"])
    subscription_query = get_subscriptions_query_with_relations(
        filter_by=[{"field": "subscriptionId", "value": subscription_id}],
        depends_on_filter=depends_on_filter,
        depends_on_subscription_filter=depends_on_subscription_filter,
    )

    response = test_client.post(
        "/api/graphql", content=subscription_query, headers={"Content-Type": "application/json"}
    )

    expected_relation_ids = [all_ids[sub_id] for sub_id in subscription_ids]

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]

    assert "errors" not in result
    assert subscriptions[0]["subscriptionId"] == subscription_id
    result_related_ids = [
        subscription["subscriptionId"] for subscription in subscriptions[0]["dependsOnSubscriptions"]["page"]
    ]
    assert_result_ids_against_expected_ids(result_related_ids, expected_relation_ids)
    result_related_statuses = {
        subscription["status"] for subscription in subscriptions[0]["dependsOnSubscriptions"]["page"]
    }
    assert result_related_statuses == subscription_statuses



import json
from http import HTTPStatus
from unittest.mock import patch

import pytest
from fastapi import Response

from orchestrator import app_settings
from test.unit_tests.helpers import assert_no_diff


def get_product_blocks_query(
    first: int = 10,
    after: int = 0,
    filter_by: list[str] | None = None,
    sort_by: list[dict[str, str]] | None = None,
    query_string: str | None = None,
) -> bytes:
    query = """
query ProductBlocksQuery($first: Int!, $after: Int!, $filterBy: [GraphqlFilter!], $sortBy: [GraphqlSort!], $query: String) {
  productBlocks(first: $first, after: $after, filterBy: $filterBy, sortBy: $sortBy, query: $query) {
    page {
      name
      endDate
      description
      createdAt
      status
      tag
      resourceTypes {
        resourceType
        description
      }
      dependsOn {
        name
        description
      }
      inUseBy {
        name
        description
      }
    }
    pageInfo {
      endCursor
      hasNextPage
      hasPreviousPage
      startCursor
      totalItems
    }
  }
}
    """
    return json.dumps(
        {
            "operationName": "ProductBlocksQuery",
            "query": query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
                "query": query_string,
            },
        }
    ).encode("utf-8")


@pytest.mark.parametrize(
    "query_args,num_results",
    [
        ({"query_string": None}, 6),
        ({"query_string": "pb1"}, 1),
        ({"query_string": "test"}, 3),
        ({"query_string": "tag:PB_*"}, 3),
        ({"query_string": "tag:pb3"}, 1),
    ],
)
def test_product_blocks_query(test_client, query_args, num_results):
    data = get_product_blocks_query(**query_args)
    response: Response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    product_blocks_data = result["data"]["productBlocks"]
    product_blocks = product_blocks_data["page"]
    pageinfo = product_blocks_data["pageInfo"]

    assert len(product_blocks) == num_results

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": num_results - 1,
        "totalItems": num_results,
    }


def test_product_blocks_payload(test_client):
    data = get_product_blocks_query(first=2, sort_by=[{"field": "name", "order": "ASC"}])
    response: Response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    product_blocks_data = result["data"]["productBlocks"]
    product_blocks = product_blocks_data["page"]
    pageinfo = product_blocks_data["pageInfo"]

    assert len(product_blocks) == 2

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": True,
        "startCursor": 0,
        "endCursor": 1,
        "totalItems": 6,
    }

    expected = [
        {
            "name": "PB_1",
            "endDate": None,
            "description": "Generic Product Block 1",
            "createdAt": "2023-05-24T00:00:00+00:00",
            "status": "ACTIVE",
            "tag": "PB1",
            "resourceTypes": [{"resourceType": "rt_1", "description": "Resource Type one"}],
            "dependsOn": [],
            "inUseBy": [],
        },
        {
            "name": "PB_2",
            "endDate": None,
            "description": "Generic Product Block 2",
            "createdAt": "2023-05-24T00:00:00+00:00",
            "status": "ACTIVE",
            "tag": "PB2",
            "resourceTypes": [
                {"resourceType": "rt_2", "description": "Resource Type two"},
                {"resourceType": "rt_3", "description": "Resource Type three"},
            ],
            "dependsOn": [],
            "inUseBy": [],
        },
    ]
    product_blocks.sort(
        key=lambda x: x["name"]
    )  # No sort in the query; sort before the `assert` to prevent flaky tests
    assert product_blocks == expected


@pytest.mark.parametrize(
    "query_args",
    [
        {"filter_by": [{"field": "name", "value": "ForTest"}]},
        {"query_string": "name:(SubBlock* | ProductBlock*)"},
    ],
)
def test_product_block_query_with_relations(test_client, query_args):
    with patch.object(app_settings, "FILTER_BY_MODE", "partial"):
        data = get_product_blocks_query(**query_args)
        response: Response = test_client.post(
            "/api/graphql", content=data, headers={"Content-Type": "application/json"}
        )

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    product_blocks_data = result["data"]["productBlocks"]
    product_blocks = product_blocks_data["page"]
    pageinfo = product_blocks_data["pageInfo"]

    assert len(product_blocks) == 3
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 2,
        "totalItems": 3,
    }

    expected = [
        {
            "name": "SubBlockOneForTest",
            "endDate": None,
            "description": "Test Sub Block One",
            "status": "ACTIVE",
            "tag": "TEST",
            "resourceTypes": [
                {"resourceType": "int_field", "description": ""},
                {"resourceType": "str_field", "description": ""},
            ],
            "dependsOn": [],
            "inUseBy": [{"name": "ProductBlockWithListUnionForTest", "description": "Test Union Sub Block"}],
        },
        {
            "name": "SubBlockTwoForTest",
            "endDate": None,
            "description": "Test Sub Block Two",
            "status": "ACTIVE",
            "tag": "TEST",
            "resourceTypes": [{"resourceType": "int_field_2", "description": ""}],
            "dependsOn": [],
            "inUseBy": [{"name": "ProductBlockWithListUnionForTest", "description": "Test Union Sub Block"}],
        },
        {
            "name": "ProductBlockWithListUnionForTest",
            "endDate": None,
            "description": "Test Union Sub Block",
            "status": "ACTIVE",
            "tag": "TEST",
            "resourceTypes": [
                {"resourceType": "int_field", "description": ""},
                {"resourceType": "str_field", "description": ""},
                {"resourceType": "list_field", "description": ""},
            ],
            "dependsOn": [
                {"name": "SubBlockOneForTest", "description": "Test Sub Block One"},
                {"name": "SubBlockTwoForTest", "description": "Test Sub Block Two"},
            ],
            "inUseBy": [],
        },
    ]

    exclude_paths = [f"root[{i}]['createdAt']" for i in range(len(expected))]
    assert_no_diff(expected, product_blocks, exclude_paths=exclude_paths)


def test_product_blocks_has_previous_page(test_client):
    data = get_product_blocks_query(after=1, sort_by=[{"field": "name", "order": "ASC"}])
    response: Response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    product_blocks_data = result["data"]["productBlocks"]
    product_blocks = product_blocks_data["page"]
    pageinfo = product_blocks_data["pageInfo"]

    assert pageinfo == {
        "hasNextPage": False,
        "hasPreviousPage": True,
        "startCursor": 1,
        "endCursor": 5,
        "totalItems": 6,
    }

    assert len(product_blocks) == 5
    product_blocks.sort(key=lambda x: x["name"])  # No sort in the query; sort before the assert to prevent flaky tests
    assert product_blocks[0]["name"] == "PB_2"
    assert product_blocks[1]["name"] == "PB_3"


@pytest.mark.parametrize(
    "query_args",
    [
        {"filter_by": [{"field": "resourceType", "value": "rt_1"}]},
        {"query_string": "resourceType:rt_1"},
    ],
)
def test_product_blocks_filter_by_resource_types(test_client, query_args):
    data = get_product_blocks_query(**query_args, sort_by=[{"field": "name", "order": "ASC"}])
    response: Response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})
    assert HTTPStatus.OK == response.status_code
    result = response.json()
    product_blocks_data = result["data"]["productBlocks"]
    product_blocks = product_blocks_data["page"]
    pageinfo = product_blocks_data["pageInfo"]

    assert pageinfo == {
        "endCursor": 0,
        "hasNextPage": False,
        "hasPreviousPage": False,
        "startCursor": 0,
        "totalItems": 1,
    }
    assert product_blocks[0]["name"] == "PB_1"


def test_product_blocks_filter_by_products(test_client):
    data = get_product_blocks_query(
        filter_by=[{"field": "product", "value": "Product 1|Product 3"}],
        sort_by=[{"field": "name", "order": "ASC"}],
    )
    response: Response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})
    assert HTTPStatus.OK == response.status_code
    result = response.json()
    product_blocks_data = result["data"]["productBlocks"]
    product_blocks = product_blocks_data["page"]
    pageinfo = product_blocks_data["pageInfo"]

    assert pageinfo == {
        "hasNextPage": False,
        "hasPreviousPage": False,
        "startCursor": 0,
        "endCursor": 1,
        "totalItems": 2,
    }
    assert product_blocks[0]["name"] == "PB_1"
    assert product_blocks[1]["name"] == "PB_2"


def test_product_blocks_sort_by_tag(test_client):
    data = get_product_blocks_query(sort_by=[{"field": "tag", "order": "DESC"}])
    response: Response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})
    assert HTTPStatus.OK == response.status_code
    result = response.json()

    product_blocks_data = result["data"]["productBlocks"]
    product_blocks = product_blocks_data["page"]
    pageinfo = product_blocks_data["pageInfo"]

    assert [p_block["tag"] for p_block in product_blocks] == ["TEST", "TEST", "TEST", "PB3", "PB2", "PB1"]
    assert pageinfo == {
        "hasNextPage": False,
        "hasPreviousPage": False,
        "startCursor": 0,
        "endCursor": 5,
        "totalItems": 6,
    }



from enum import StrEnum
from ipaddress import IPv4Address
from typing import Annotated

from annotated_types import Ge, Le, Len
from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SI, SubscriptionLifecycle
from pydantic import computed_field

from products.product_blocks.example2 import Example2Block, Example2BlockInactive, Example2BlockProvisioning


class ExampleStrEnum1(StrEnum):
    option1 = "option1"
    option2 = "option2"
    option3 = "option3"


ListOfEight = Annotated[list[SI], Len(min_length=1, max_length=8)]


AnnotatedInt = Annotated[int, Ge(1), Le(32_767)]


class Example1BlockInactive(ProductBlockModel, product_block_name="Example1"):
    example_str_enum_1: ExampleStrEnum1 = ExampleStrEnum1.option2
    example2: Example2BlockInactive | None = None
    unmodifiable_str: str | None = None
    eight: ListOfEight[Example2BlockInactive]
    modifiable_boolean: bool | None = False
    annotated_int: AnnotatedInt | None = None
    imported_type: IPv4Address | None = None
    always_optional_str: str | None = None


class Example1BlockProvisioning(Example1BlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
    example_str_enum_1: ExampleStrEnum1
    example2: Example2BlockProvisioning
    unmodifiable_str: str
    eight: ListOfEight[Example2BlockProvisioning]
    modifiable_boolean: bool
    annotated_int: AnnotatedInt | None = None
    imported_type: IPv4Address | None = None
    always_optional_str: str | None = None

    @computed_field
    @property
    def title(self) -> str:
        # TODO: format correct title string
        return f"{self.name}"


class Example1Block(Example1BlockProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    example_str_enum_1: ExampleStrEnum1
    example2: Example2Block
    unmodifiable_str: str
    eight: ListOfEight[Example2Block]
    modifiable_boolean: bool
    annotated_int: AnnotatedInt
    imported_type: IPv4Address
    always_optional_str: str | None = None



# Copyright 2022 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import datetime
import json
from http import HTTPStatus
from unittest.mock import patch

import pytest
from sqlalchemy import select

from orchestrator import app_settings
from orchestrator.db import SubscriptionMetadataTable, db
from orchestrator.db.models import SubscriptionCustomerDescriptionTable, SubscriptionTable
from orchestrator.domain.base import SubscriptionModel
from orchestrator.types import SubscriptionLifecycle
from test.unit_tests.conftest import do_refresh_subscriptions_search_view

subscription_fields = [
    "endDate",
    "description",
    "subscriptionId",
    "startDate",
    "status",
    "insync",
    "note",
    "product",
    "productBlockInstances",
]
subscription_product_fields = [
    "productId",
    "name",
    "description",
    "productType",
    "status",
    "tag",
    "createdAt",
    "endDate",
]
CUSTOMER_ID = "a6274018-0580-41cd-9fe8-17b946706e9f"


def build_subscriptions_query(body: str) -> str:
    return f"""
query SubscriptionQuery(
    $first: Int!,
    $after: Int!,
    $sortBy: [GraphqlSort!],
    $filterBy: [GraphqlFilter!],
    $query: String
) {{
    subscriptions(first: $first, after: $after, sortBy: $sortBy, filterBy: $filterBy, query: $query)
    {body}
}}
"""


def get_subscriptions_query(
    first: int = 10,
    after: int = 0,
    filter_by: list[dict] | None = None,
    sort_by: list | None = None,
    query_string: str | None = None,
) -> bytes:
    gql_query = build_subscriptions_query(
        """{
    page {
      description
      subscriptionId
      status
      insync
      note
      startDate
      endDate
      productBlockInstances {
        id
        parent
        subscriptionInstanceId
        ownerSubscriptionId
        productBlockInstanceValues
        inUseByRelations
      }
      product {
        productId
        name
        description
        productType
        status
        tag
        createdAt
        endDate
      }
      customerDescriptions {
        subscriptionId
        customerId
        description
      }
    }
    pageInfo {
      startCursor
      totalItems
      hasPreviousPage
      endCursor
      hasNextPage
    }
  }
"""
    )
    return json.dumps(
        {
            "operationName": "SubscriptionQuery",
            "query": gql_query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
                "query": query_string,
            },
        }
    ).encode("utf-8")


def get_subscriptions_query_with_relations(
    first: int = 10,
    after: int = 0,
    filter_by: list[dict[str, str]] | None = None,
    sort_by: list[dict[str, str]] | None = None,
    query_string: str | None = None,
) -> bytes:
    query = build_subscriptions_query(
        """{
    page {
      description
      subscriptionId
      status
      insync
      note
      startDate
      endDate
      productBlockInstances {
        ownerSubscriptionId
        inUseByRelations
      }
      processes(sortBy: [{field: "startedAt", order: ASC}]) {
        page {
          processId
          isTask
          lastStep
          lastStatus
          assignee
          failedReason
          traceback
          workflowName
          createdBy
          startedAt
          lastModifiedAt
          product {
            productId
            name
            description
            productType
            status
            tag
            createdAt
            endDate
          }
        }
      }
      dependsOnSubscriptions {
        page {
          description
          subscriptionId
          status
          insync
          note
          startDate
          endDate
        }
      }
      inUseBySubscriptions {
        page {
          description
          subscriptionId
          status
          insync
          note
          startDate
          endDate
        }
      }
    }
    pageInfo {
      startCursor
      totalItems
      hasPreviousPage
      endCursor
      hasNextPage
    }
  }
    """
    )
    return json.dumps(
        {
            "operationName": "SubscriptionQuery",
            "query": query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
                "query": query_string,
            },
        }
    ).encode("utf-8")


def get_subscriptions_product_block_json_schema_query(
    first: int = 10,
    after: int = 0,
    filter_by: list[dict[str, str]] | None = None,
    sort_by: list[dict[str, str]] | None = None,
    query_string: str | None = None,
) -> bytes:
    query = build_subscriptions_query(
        """{
    page {
      _schema
    }
    pageInfo {
      startCursor
      totalItems
      hasPreviousPage
      endCursor
      hasNextPage
    }
  }
    """
    )
    return json.dumps(
        {
            "operationName": "SubscriptionQuery",
            "query": query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
                "query": query_string,
            },
        }
    ).encode("utf-8")


def get_subscriptions_product_generic_one(
    first: int = 10,
    after: int = 0,
    filter_by: list[str] | None = None,
    sort_by: list[dict[str, str]] | None = None,
    query_string: str | None = None,
) -> bytes:
    query = build_subscriptions_query(
        """{
    page {
      ... on GenericProductOneSubscription {
        description
        subscriptionId
        status
        insync
        note
        startDate
        endDate
        pb1 {
          rt1
        }
        pb2 {
          rt2
          rt3
        }
        customerDescriptions {
            subscriptionId
            customerId
            description
        }
      }
    }
    pageInfo {
      startCursor
      totalItems
      hasPreviousPage
      endCursor
      hasNextPage
    }
  }
    """
    )
    return json.dumps(
        {
            "operationName": "SubscriptionQuery",
            "query": query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
                "query": query_string,
            },
        }
    ).encode("utf-8")


def get_subscriptions_product_sub_list_union(
    first: int = 10,
    after: int = 0,
    filter_by: list[dict[str, str]] | None = None,
    sort_by: list[str] | None = None,
    query_string: str | None = None,
) -> bytes:
    query = build_subscriptions_query(
        """{
    page {
      ... on ProductSubListUnionSubscription {
        description
        subscriptionId
        status
        insync
        note
        startDate
        endDate
        product{
          productType
        }
        testBlock {
          intField
          strField
          listField
          listUnionBlocks {
            ... on SubBlockOneForTestInactive {
              intField
              strField
            }
            ... on SubBlockTwoForTestInactive {
              intField2
            }
          }
        }
      }
    }
    pageInfo {
      startCursor
      totalItems
      hasPreviousPage
      endCursor
      hasNextPage
    }
  }
    """
    )

    return json.dumps(
        {
            "operationName": "SubscriptionQuery",
            "query": query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
                "query": query_string,
            },
        }
    ).encode("utf-8")


def get_subscriptions_with_metadata_and_schema_query(
    first: int = 1,
    after: int = 0,
    filter_by: list[dict[str, str]] | None = None,
    sort_by: list[dict[str, str]] | None = None,
    query_string: str | None = None,
) -> bytes:
    gql_query = build_subscriptions_query(
        """{
    page {
      subscriptionId
      metadata
      _metadataSchema
    }
  }
    """
    )
    return json.dumps(
        {
            "operationName": "SubscriptionQuery",
            "query": gql_query,
            "variables": {
                "first": first,
                "after": after,
                "sortBy": sort_by if sort_by else [],
                "filterBy": filter_by if filter_by else [],
                "query": query_string,
            },
        }
    ).encode("utf-8")


def test_subscriptions_single_page(test_client, product_type_1_subscriptions_factory, benchmark, monitor_sqlalchemy):
    # when

    product_type_1_subscriptions_factory(4)
    data = get_subscriptions_query()

    with monitor_sqlalchemy():

        @benchmark
        def response():
            return test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    assert "errors" not in result

    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 6,
        "totalItems": 7,
    }

    for subscription in subscriptions:
        for field in subscription_fields:
            assert field in subscription
        for field in subscription_product_fields:
            assert field in subscription["product"]


def test_subscriptions_has_next_page(test_client, product_type_1_subscriptions_factory):
    # when

    product_type_1_subscriptions_factory(30)
    data = get_subscriptions_query()
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    assert "errors" not in result

    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": True,
        "startCursor": 0,
        "endCursor": 9,
        "totalItems": 33,
    }

    for subscription in subscriptions:
        for field in subscription_fields:
            assert field in subscription


def test_subscriptions_has_previous_page(test_client, product_type_1_subscriptions_factory):
    # when

    product_type_1_subscriptions_factory(30)
    data = get_subscriptions_query(after=1)
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    assert "errors" not in result

    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert len(subscriptions) == 10

    assert pageinfo == {
        "hasPreviousPage": True,
        "hasNextPage": True,
        "startCursor": 1,
        "endCursor": 10,
        "totalItems": 33,
    }


def test_subscriptions_sorting_asc(test_client, product_type_1_subscriptions_factory):
    # when

    product_type_1_subscriptions_factory(30)
    data = get_subscriptions_query(sort_by=[{"field": "startDate", "order": "ASC"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    assert "errors" not in result

    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": True,
        "startCursor": 0,
        "endCursor": 9,
        "totalItems": 33,
    }

    for i in range(0, 8):
        assert subscriptions[i]["startDate"] < subscriptions[i + 1]["startDate"]


def test_subscriptions_sorting_desc(test_client, product_type_1_subscriptions_factory):
    # when

    product_type_1_subscriptions_factory(30)
    data = get_subscriptions_query(sort_by=[{"field": "startDate", "order": "DESC"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    assert "errors" not in result

    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": True,
        "startCursor": 0,
        "endCursor": 9,
        "totalItems": 33,
    }

    for i in range(0, 8):
        assert subscriptions[i + 1]["startDate"] < subscriptions[i]["startDate"]


def test_subscriptions_sorting_product_tag_asc(test_client, generic_subscription_1, generic_subscription_2):
    # when

    data = get_subscriptions_query(sort_by=[{"field": "tag", "order": "ASC"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    assert "errors" not in result

    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 4,
        "totalItems": 5,
    }

    product_tag_list = [subscription["product"]["tag"] for subscription in subscriptions]
    assert product_tag_list == ["GEN1", "GEN2", "Sub", "Sub", "UnionSub"]


def test_subscriptions_sorting_product_tag_desc(test_client, generic_subscription_1, generic_subscription_2):
    # when

    data = get_subscriptions_query(sort_by=[{"field": "tag", "order": "DESC"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    assert "errors" not in result

    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 4,
        "totalItems": 5,
    }

    product_tag_list = [subscription["product"]["tag"] for subscription in subscriptions]
    assert product_tag_list == ["UnionSub", "Sub", "Sub", "GEN2", "GEN1"]


def test_subscriptions_sorting_invalid_field(test_client, product_type_1_subscriptions_factory):
    # when

    product_type_1_subscriptions_factory(30)
    data = get_subscriptions_query(sort_by=[{"field": "start_date", "order": "DESC"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    pageinfo = subscriptions_data["pageInfo"]

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": True,
        "startCursor": 0,
        "endCursor": 9,
        "totalItems": 33,
    }

    assert "errors" in result
    assert result["errors"] == [
        {
            "message": (
                "Invalid sort arguments (invalid_sorting=['start_date'] "
                "valid_sort_keys=['customerId', 'description', 'endDate', "
                "'insync', 'note', 'productCreatedAt', 'productDescription', "
                "'productEndDate', 'productId', 'productName', 'productStatus', "
                "'productTag', 'productType', 'startDate', 'status', "
                "'subscriptionId', 'tag', 'version'])"
            ),
            "path": ["subscriptions"],
            "extensions": {"error_type": "bad_request"},
        }
    ]


def test_subscriptions_sorting_invalid_order(test_client, product_type_1_subscriptions_factory):
    # when

    product_type_1_subscriptions_factory(30)
    data = get_subscriptions_query(sort_by=[{"field": "start_date", "order": "test"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()

    assert not result["data"]
    assert "errors" in result
    assert "Internal Server Error" in result["errors"][0]["message"]


@pytest.mark.parametrize(
    "query_args",
    [
        {"filter_by": [{"field": "status", "value": SubscriptionLifecycle.TERMINATED}]},
        {"query_string": "status:terminated"},
    ],
)
def test_subscriptions_filtering_on_status(
    test_client, product_type_1_subscriptions_factory, generic_product_type_1, query_args
):
    # when

    subscription_ids = product_type_1_subscriptions_factory(30)

    GenericProductOneInactive, GenericProductOne = generic_product_type_1
    subscription_1 = GenericProductOne.from_subscription(subscription_ids[0])
    subscription_1 = subscription_1.from_other_lifecycle(subscription_1, SubscriptionLifecycle.TERMINATED)
    subscription_1.save()
    subscription_9 = GenericProductOne.from_subscription(subscription_ids[8])
    subscription_9 = subscription_9.from_other_lifecycle(subscription_9, SubscriptionLifecycle.TERMINATED)
    subscription_9.save()
    do_refresh_subscriptions_search_view()

    subscription_query = get_subscriptions_query(**query_args)

    response = test_client.post(
        "/api/graphql", content=subscription_query, headers={"Content-Type": "application/json"}
    )

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]

    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result

    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 1,
        "totalItems": 2,
    }

    result_subscription_ids = {subscription["subscriptionId"] for subscription in subscriptions}
    assert result_subscription_ids == {str(subscription_1.subscription_id), str(subscription_9.subscription_id)}
    assert subscriptions[0]["status"] == "TERMINATED"
    assert subscriptions[1]["status"] == "TERMINATED"


def test_subscriptions_range_filtering_on_start_date(test_client, product_type_1_subscriptions_factory):
    # when

    product_type_1_subscriptions_factory(30)

    with patch.object(app_settings, "FILTER_BY_MODE", "partial"):
        data = get_subscriptions_query(first=1, filter_by=[{"field": "startDate", "value": "2023-05-24"}])
        response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})
        first_subscription = response.json()["data"]["subscriptions"]["page"][0]

    first_subscription_date = datetime.datetime.fromisoformat(first_subscription["startDate"])
    higher_than_date = first_subscription_date.isoformat()
    lower_than_date = (first_subscription_date + datetime.timedelta(days=3)).isoformat()

    data = get_subscriptions_query(
        filter_by=[
            {"field": "startDate", "value": f">={higher_than_date}"},
            {"field": "startDate", "value": f"<={lower_than_date}"},
        ]
    )
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 3,
        "totalItems": 4,
    }

    for subscription in subscriptions:
        assert higher_than_date <= subscription["startDate"] <= lower_than_date


def test_subscriptions_with_exact_filter_by(test_client, product_type_1_subscriptions_factory):
    # when

    product_type_1_subscriptions_factory(20)

    with patch.object(app_settings, "FILTER_BY_MODE", "exact"):
        data = get_subscriptions_query(filter_by=[{"field": "description", "value": "Subscription 1"}])
        response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert len(subscriptions) == 1
    assert "errors" not in result
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": 1,
    }


def test_subscriptions_range_filtering_on_type(test_client, product_type_1_subscriptions_factory):
    # when

    product_type_1_subscriptions_factory(15)

    data = get_subscriptions_query(filter_by=[{"field": "type", "value": "Test"}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 2,
        "totalItems": 3,
    }

    for subscription in subscriptions:
        assert subscription["product"]["productType"] == "Test"


def test_subscriptions_filtering_with_invalid_filter(
    test_client, product_type_1_subscriptions_factory, generic_product_type_1
):
    # when

    subscription_ids = product_type_1_subscriptions_factory(30)

    GenericProductOneInactive, GenericProductOne = generic_product_type_1
    subscription_1 = GenericProductOne.from_subscription(subscription_ids[0])
    subscription_1 = subscription_1.from_other_lifecycle(subscription_1, SubscriptionLifecycle.TERMINATED)
    subscription_1.save()
    subscription_9 = GenericProductOne.from_subscription(subscription_ids[8])
    subscription_9 = subscription_9.from_other_lifecycle(subscription_9, SubscriptionLifecycle.TERMINATED)
    subscription_9.save()

    data = get_subscriptions_query(
        filter_by=[
            {"field": "status", "value": SubscriptionLifecycle.TERMINATED},
            {"field": "test", "value": "invalid"},
        ]
    )
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    errors = result["errors"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert errors == [
        {
            "message": (
                "Invalid filter arguments (invalid_filters=['test'] "
                "valid_filter_keys=['customerId', 'description', 'endDate', 'insync', "
                "'note', 'product', 'productId', 'startDate', 'status', 'subscriptionId', 'tag', 'type', 'version'])"
            ),
            "path": ["subscriptions"],
            "extensions": {"error_type": "bad_request"},
        }
    ]
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 1,
        "totalItems": 2,
    }

    for subscription in subscriptions:
        assert subscription["status"] == "TERMINATED"


@pytest.mark.parametrize(
    "query_args",
    [
        lambda sid: {"filter_by": [{"field": "subscriptionId", "value": sid}]},
        lambda sid: {"query_string": f"subscription_id:{sid}"},
        lambda sid: {"query_string": f"subscriptionId:{sid}"},
        lambda sid: {"query_string": f"{sid}"},
        lambda sid: {"query_string": f"{sid.split('-')[0]}"},
        lambda sid: {"filter_by": [{"field": "customerId", "value": CUSTOMER_ID.split("-")[0]}]},
        lambda sid: {"query_string": CUSTOMER_ID.split("-")[0]},
    ],
)
def test_single_subscription(test_client, product_type_1_subscriptions_factory, generic_product_type_1, query_args):
    # given

    _, GenericProductOne = generic_product_type_1
    subscription_ids = product_type_1_subscriptions_factory(30)

    do_refresh_subscriptions_search_view()

    subscription_id = subscription_ids[10]
    subscription = db.session.execute(
        select(SubscriptionTable).filter(SubscriptionTable.subscription_id == subscription_id)
    ).scalar_one_or_none()
    subscription.customer_id = CUSTOMER_ID
    subscription.customer_descriptions = [
        SubscriptionCustomerDescriptionTable(
            subscription_id=subscription_id, customer_id=CUSTOMER_ID, description="customer alias"
        )
    ]
    db.session.add(subscription)
    db.session.commit()

    do_refresh_subscriptions_search_view()

    # when
    with patch.object(app_settings, "FILTER_BY_MODE", "partial"):
        data = get_subscriptions_query(**query_args(subscription_id))
        response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    subscription = GenericProductOne.from_subscription(subscription_id)

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert len(subscriptions) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": 1,
    }
    assert subscriptions[0]["subscriptionId"] == subscription_id
    assert subscriptions[0]["status"] == SubscriptionLifecycle.ACTIVE.name
    assert subscriptions[0]["customerDescriptions"] == [
        {
            "subscriptionId": subscription_id,
            "customerId": "a6274018-0580-41cd-9fe8-17b946706e9f",
            "description": "customer alias",
        }
    ]
    assert subscriptions[0]["productBlockInstances"] == [
        {
            "id": 0,
            "subscriptionInstanceId": str(subscription.pb_1.subscription_instance_id),
            "ownerSubscriptionId": subscription_id,
            "parent": None,
            "productBlockInstanceValues": [
                {"field": "name", "value": "PB_1"},
                {"field": "label", "value": None},
                {"field": "rt1", "value": "Value1"},
            ],
            "inUseByRelations": [],
        },
        {
            "id": 1,
            "subscriptionInstanceId": str(subscription.pb_2.subscription_instance_id),
            "ownerSubscriptionId": subscription_id,
            "parent": None,
            "productBlockInstanceValues": [
                {"field": "name", "value": "PB_2"},
                {"field": "label", "value": None},
                {"field": "rt2", "value": 42},
                {"field": "rt3", "value": "Value2"},
            ],
            "inUseByRelations": [],
        },
    ]


@pytest.mark.parametrize(
    "query_args,total_items",
    [
        (lambda ids: {}, 33),
        (lambda ids: {"filter_by": [{"field": "subscriptionId", "value": f"{ids[0]}|{ids[1]}"}]}, 2),
        (lambda ids: {"filter_by": [{"field": "subscriptionId", "value": "|".join(ids[3:8])}]}, 5),
        (lambda ids: {"query_string": f"subscriptionId:({ids[2]}|{ids[3]})"}, 2),
    ],
)
def test_multiple_subscriptions(
    test_client, product_type_1_subscriptions_factory, generic_product_type_1, query_args, total_items
):
    # given

    _, GenericProductOne = generic_product_type_1
    subscription_ids = product_type_1_subscriptions_factory(30)

    do_refresh_subscriptions_search_view()

    # when
    data = get_subscriptions_query(first=100, **query_args(subscription_ids))
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert len(subscriptions) == total_items
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": total_items - 1,
        "totalItems": total_items,
    }


@pytest.mark.parametrize(
    "query_args",
    [
        lambda sid: {"filter_by": [{"field": "subscriptionId", "value": sid}]},
        lambda sid: {"query_string": sid},
    ],
)
def test_single_subscription_with_processes(
    fastapi_app_graphql,
    test_client,
    product_type_1_subscriptions_factory,
    mocked_processes,
    mocked_processes_resumeall,  # noqa: F811
    generic_subscription_2,  # noqa: F811
    generic_subscription_1,
    query_args,
):
    # when

    product_type_1_subscriptions_factory(30)
    subscription_id = generic_subscription_1

    do_refresh_subscriptions_search_view()

    data = get_subscriptions_query_with_relations(**query_args(subscription_id))
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert len(subscriptions) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": 1,
    }
    assert subscriptions[0]["subscriptionId"] == subscription_id
    assert subscriptions[0]["processes"]["page"][0]["processId"] == str(mocked_processes[0])


def test_single_subscription_with_depends_on_subscriptions(
    fastapi_app_graphql,
    test_client,
    product_type_1_subscriptions_factory,
    sub_one_subscription_1,
    sub_two_subscription_1,
    product_sub_list_union_subscription_1,
    benchmark,
):
    # when

    product_type_1_subscriptions_factory(30)

    do_refresh_subscriptions_search_view()

    subscription_id = str(product_sub_list_union_subscription_1)
    data = get_subscriptions_query_with_relations(query_string=subscription_id)

    @benchmark
    def response():
        return test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    expected_depends_on_ids = {
        str(subscription.subscription_id) for subscription in [sub_one_subscription_1, sub_two_subscription_1]
    }
    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert len(subscriptions) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": 1,
    }
    assert subscriptions[0]["subscriptionId"] == subscription_id
    assert len(subscriptions[0]["processes"]["page"]) == 0
    depends_on_ids = subscriptions[0]["dependsOnSubscriptions"]["page"]
    result_depends_on_ids = {subscription["subscriptionId"] for subscription in depends_on_ids}
    assert result_depends_on_ids == expected_depends_on_ids
    assert len(subscriptions[0]["inUseBySubscriptions"]["page"]) == 0


def test_single_subscription_with_in_use_by_subscriptions(
    fastapi_app_graphql,
    test_client,
    product_type_1_subscriptions_factory,
    sub_one_subscription_1,
    product_sub_list_union_subscription_1,
):
    # when

    product_type_1_subscriptions_factory(30)

    subscription_id = str(sub_one_subscription_1.subscription_id)
    subscription_query = get_subscriptions_query_with_relations(
        filter_by=[{"field": "subscriptionId", "value": subscription_id}]
    )

    response = test_client.post(
        "/api/graphql", content=subscription_query, headers={"Content-Type": "application/json"}
    )

    expected_in_use_by_ids = [str(product_sub_list_union_subscription_1)]

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert "errors" not in result
    assert len(subscriptions) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": 1,
    }
    assert subscriptions[0]["subscriptionId"] == subscription_id
    assert len(subscriptions[0]["processes"]["page"]) == 0
    assert len(subscriptions[0]["dependsOnSubscriptions"]["page"]) == 0
    result_in_use_by_ids = [
        subscription["subscriptionId"] for subscription in subscriptions[0]["inUseBySubscriptions"]["page"]
    ]
    assert result_in_use_by_ids == expected_in_use_by_ids

    list_sub = SubscriptionModel.from_subscription(product_sub_list_union_subscription_1)
    assert subscriptions[0]["productBlockInstances"] == [
        {
            "ownerSubscriptionId": subscription_id,
            "inUseByRelations": [
                {
                    "subscription_id": str(product_sub_list_union_subscription_1),
                    "subscription_instance_id": str(list_sub.test_block.subscription_instance_id),
                }
            ],
        }
    ]


def test_single_subscription_schema(
    fastapi_app_graphql,
    test_client,
    product_type_1_subscriptions_factory,
    sub_one_subscription_1,
    sub_two_subscription_1,
    product_sub_list_union_subscription_1,
    benchmark,
):
    # when

    product_type_1_subscriptions_factory(30)
    subscription_id = str(product_sub_list_union_subscription_1)
    data = get_subscriptions_product_block_json_schema_query(
        filter_by=[{"field": "subscriptionId", "value": subscription_id}]
    )

    @benchmark
    def response():
        return test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]

    assert "errors" not in result
    expected = {
        "$defs": {
            "ProductBlockWithListUnionForTestInactive": {
                "description": "Valid for statuses: all others\n\nSee `active` version.",
                "properties": {
                    "name": {"anyOf": [{"type": "string"}, {"type": "null"}], "title": "Name"},
                    "subscription_instance_id": {
                        "format": "uuid",
                        "title": "Subscription Instance Id",
                        "type": "string",
                    },
                    "owner_subscription_id": {"format": "uuid", "title": "Owner Subscription Id", "type": "string"},
                    "label": {"anyOf": [{"type": "string"}, {"type": "null"}], "default": None, "title": "Label"},
                    "list_union_blocks": {
                        "items": {
                            "anyOf": [
                                {"$ref": "#/$defs/SubBlockTwoForTestInactive"},
                                {"$ref": "#/$defs/SubBlockOneForTestInactive"},
                            ]
                        },
                        "title": "List Union Blocks",
                        "type": "array",
                    },
                    "int_field": {
                        "anyOf": [{"type": "integer"}, {"type": "null"}],
                        "default": None,
                        "title": "Int Field",
                    },
                    "str_field": {
                        "anyOf": [{"type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "Str Field",
                    },
                    "list_field": {"items": {"type": "integer"}, "title": "List Field", "type": "array"},
                },
                "required": ["name", "subscription_instance_id", "owner_subscription_id", "list_union_blocks"],
                "title": "ProductBlockWithListUnionForTestInactive",
                "type": "object",
            },
            "ProductLifecycle": {
                "enum": ["active", "pre production", "phase out", "end of life"],
                "title": "ProductLifecycle",
                "type": "string",
            },
            "ProductModel": {
                "description": "Represent the product as defined in the database as a dataclass.",
                "properties": {
                    "product_id": {"format": "uuid", "title": "Product Id", "type": "string"},
                    "name": {"title": "Name", "type": "string"},
                    "description": {"title": "Description", "type": "string"},
                    "product_type": {"title": "Product Type", "type": "string"},
                    "tag": {"title": "Tag", "type": "string"},
                    "status": {"$ref": "#/$defs/ProductLifecycle"},
                    "created_at": {
                        "anyOf": [{"format": "date-time", "type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "Created At",
                    },
                    "end_date": {
                        "anyOf": [{"format": "date-time", "type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "End Date",
                    },
                },
                "required": ["product_id", "name", "description", "product_type", "tag", "status"],
                "title": "ProductModel",
                "type": "object",
            },
            "SubBlockOneForTestInactive": {
                "description": "Valid for statuses: all others\n\nSee `active` version.",
                "properties": {
                    "name": {"anyOf": [{"type": "string"}, {"type": "null"}], "title": "Name"},
                    "subscription_instance_id": {
                        "format": "uuid",
                        "title": "Subscription Instance Id",
                        "type": "string",
                    },
                    "owner_subscription_id": {"format": "uuid", "title": "Owner Subscription Id", "type": "string"},
                    "label": {"anyOf": [{"type": "string"}, {"type": "null"}], "default": None, "title": "Label"},
                    "int_field": {
                        "anyOf": [{"type": "integer"}, {"type": "null"}],
                        "default": None,
                        "title": "Int Field",
                    },
                    "str_field": {
                        "anyOf": [{"type": "string"}, {"type": "null"}],
                        "default": None,
                        "title": "Str Field",
                    },
                },
                "required": ["name", "subscription_instance_id", "owner_subscription_id"],
                "title": "SubBlockOneForTestInactive",
                "type": "object",
            },
            "SubBlockTwoForTestInactive": {
                "description": "Valid for statuses: all others\n\nSee `active` version.",
                "properties": {
                    "name": {"anyOf": [{"type": "string"}, {"type": "null"}], "title": "Name"},
                    "subscription_instance_id": {
                        "format": "uuid",
                        "title": "Subscription Instance Id",
                        "type": "string",
                    },
                    "owner_subscription_id": {"format": "uuid", "title": "Owner Subscription Id", "type": "string"},
                    "label": {"anyOf": [{"type": "string"}, {"type": "null"}], "default": None, "title": "Label"},
                    "int_field_2": {"title": "Int Field 2", "type": "integer"},
                },
                "required": ["name", "subscription_instance_id", "owner_subscription_id", "int_field_2"],
                "title": "SubBlockTwoForTestInactive",
                "type": "object",
            },
            "SubscriptionLifecycle": {
                "enum": ["initial", "active", "migrating", "disabled", "terminated", "provisioning"],
                "title": "SubscriptionLifecycle",
                "type": "string",
            },
        },
        "description": "Valid for statuses: all others\n\nSee `active` version.",
        "properties": {
            "product": {"$ref": "#/$defs/ProductModel"},
            "customer_id": {"title": "Customer Id", "type": "string"},
            "subscription_id": {"format": "uuid", "title": "Subscription Id", "type": "string"},
            "description": {"default": "Initial subscription", "title": "Description", "type": "string"},
            "status": {"allOf": [{"$ref": "#/$defs/SubscriptionLifecycle"}], "default": "initial"},
            "insync": {"default": False, "title": "Insync", "type": "boolean"},
            "start_date": {
                "anyOf": [{"format": "date-time", "type": "string"}, {"type": "null"}],
                "default": None,
                "title": "Start Date",
            },
            "end_date": {
                "anyOf": [{"format": "date-time", "type": "string"}, {"type": "null"}],
                "default": None,
                "title": "End Date",
            },
            "note": {"anyOf": [{"type": "string"}, {"type": "null"}], "default": None, "title": "Note"},
            "test_block": {"anyOf": [{"$ref": "#/$defs/ProductBlockWithListUnionForTestInactive"}, {"type": "null"}]},
            "version": {
                "default": 1,
                "title": "Version",
                "type": "integer",
            },
        },
        "required": ["product", "customer_id", "test_block"],
        "title": "ProductSubListUnionInactive",
        "type": "object",
    }
    actual = subscriptions[0]["_schema"]
    assert actual == expected


def test_single_subscription_metadata_and_schema(
    fastapi_app_graphql,
    test_client,
    sub_one_subscription_1,
):
    # when
    expected_metadata = {"some_metadata_prop": ["test value 1", "test 2"]}
    subscription_id = str(sub_one_subscription_1.subscription_id)
    subscription_metadata = SubscriptionMetadataTable(subscription_id=subscription_id, metadata_=expected_metadata)
    db.session.add(subscription_metadata)
    db.session.commit()

    data = get_subscriptions_with_metadata_and_schema_query(
        filter_by=[{"field": "subscriptionId", "value": subscription_id}]
    )
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})
    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()
    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]

    assert "errors" not in result
    assert subscriptions[0]["subscriptionId"] == subscription_id
    assert subscriptions[0]["metadata"] == expected_metadata
    assert subscriptions[0]["_metadataSchema"] == {
        "title": "Metadata",
        "type": "object",
        "properties": {
            "some_metadata_prop": {"title": "Some Metadata Prop", "type": "array", "items": {"type": "string"}}
        },
        "required": ["some_metadata_prop"],
    }


def test_subscriptions_product_generic_one(
    fastapi_app_graphql,
    test_client,
    product_type_1_subscriptions_factory,
):
    # when

    subscriptions = product_type_1_subscriptions_factory(30)
    subscription_id = str(subscriptions[0])
    subscription = db.session.execute(
        select(SubscriptionTable).filter(SubscriptionTable.subscription_id == subscription_id)
    ).scalar_one_or_none()
    subscription.customer_descriptions = [
        SubscriptionCustomerDescriptionTable(customer_id=CUSTOMER_ID, description="customer alias")
    ]
    db.session.add(subscription)
    db.session.commit()

    data = get_subscriptions_product_generic_one(filter_by=[{"field": "subscriptionId", "value": subscription_id}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()

    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert len(subscriptions) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": 1,
    }
    assert subscriptions[0]["subscriptionId"] == subscription_id
    assert subscriptions[0]["status"] == SubscriptionLifecycle.ACTIVE.name
    assert subscriptions[0]["pb1"] == {"rt1": "Value1"}
    assert subscriptions[0]["pb2"] == {"rt2": 42, "rt3": "Value2"}
    assert subscriptions[0]["customerDescriptions"] == [
        {
            "subscriptionId": subscription_id,
            "customerId": "a6274018-0580-41cd-9fe8-17b946706e9f",
            "description": "customer alias",
        }
    ]


def test_single_subscription_product_list_union_type(
    fastapi_app_graphql,
    test_client,
    product_sub_list_union_subscription_1,
):
    # when

    subscription_id = str(product_sub_list_union_subscription_1)
    data = get_subscriptions_product_sub_list_union(filter_by=[{"field": "subscriptionId", "value": subscription_id}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()

    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert len(subscriptions) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": 1,
    }
    assert subscriptions[0]["subscriptionId"] == subscription_id
    assert subscriptions[0]["status"] == SubscriptionLifecycle.ACTIVE.name
    assert subscriptions[0]["testBlock"] == {
        "intField": 1,
        "strField": "blah",
        "listField": [2],
        "listUnionBlocks": [
            {"intField": 1, "strField": "blah"},
            {"intField2": 3},
        ],
    }


def test_single_subscription_product_list_union_type_provisioning_subscription(
    fastapi_app_graphql,
    test_client,
    product_sub_list_union_subscription_1,
):
    # when

    subscription = SubscriptionModel.from_subscription(product_sub_list_union_subscription_1)
    subscription = subscription.from_other_lifecycle(subscription, SubscriptionLifecycle.PROVISIONING)
    subscription.save()

    subscription_id = str(product_sub_list_union_subscription_1)
    data = get_subscriptions_product_sub_list_union(filter_by=[{"field": "subscriptionId", "value": subscription_id}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()

    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert len(subscriptions) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": 1,
    }
    assert subscriptions[0]["subscriptionId"] == subscription_id
    assert subscriptions[0]["testBlock"] == {
        "intField": 1,
        "strField": "blah",
        "listField": [2],
        "listUnionBlocks": [
            {"intField": 1, "strField": "blah"},
            {"intField2": 3},
        ],
    }


def test_single_subscription_product_list_union_type_terminated_subscription(
    fastapi_app_graphql,
    test_client,
    product_sub_list_union_subscription_1,
):
    # when

    subscription = SubscriptionModel.from_subscription(product_sub_list_union_subscription_1)
    subscription = subscription.from_other_lifecycle(subscription, SubscriptionLifecycle.TERMINATED)
    subscription.save()

    subscription_id = str(product_sub_list_union_subscription_1)
    data = get_subscriptions_product_sub_list_union(filter_by=[{"field": "subscriptionId", "value": subscription_id}])
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})

    # then

    assert HTTPStatus.OK == response.status_code
    result = response.json()

    subscriptions_data = result["data"]["subscriptions"]
    subscriptions = subscriptions_data["page"]
    pageinfo = subscriptions_data["pageInfo"]

    assert len(subscriptions) == 1
    assert pageinfo == {
        "hasPreviousPage": False,
        "hasNextPage": False,
        "startCursor": 0,
        "endCursor": 0,
        "totalItems": 1,
    }
    assert subscriptions[0]["subscriptionId"] == subscription_id
    assert subscriptions[0]["testBlock"] == {
        "intField": 1,
        "strField": "blah",
        "listField": [2],
        "listUnionBlocks": [
            {"intField": 1, "strField": "blah"},
            {"intField2": 3},
        ],
    }



from orchestrator.forms.validators import CustomerId
from pydantic_forms.core import FormPage


def test_customer_id_schema():
    class Form(FormPage):
        customer_id: CustomerId

    expected = {
        "additionalProperties": False,
        "properties": {"customer_id": {"format": "customerId", "title": "Customer Id", "type": "string"}},
        "required": ["customer_id"],
        "title": "unknown",
        "type": "object",
    }

    assert Form.model_json_schema() == expected



from unittest import mock

from pydantic import computed_field

from orchestrator.domain import SubscriptionModel
from orchestrator.domain.base import DomainModel
from orchestrator.domain.context_cache import cache_subscription_models


def test_serializable_property():
    class DerivedDomainModel(DomainModel):
        @computed_field  # type: ignore[misc]
        @property
        def double_int_field(self) -> int:
            # This property is serialized
            return 2 * self.int_field

        @property
        def triple_int_field(self) -> int:
            # This property is not serialized
            return 3 * self.int_field

        int_field: int

    block = DerivedDomainModel(int_field=13)

    assert block.model_dump() == {"int_field": 13, "double_int_field": 26}


def test_inherited_serializable_property():
    class ProvisioningDomainModel(DomainModel):
        @computed_field  # type: ignore[misc]
        @property
        def double_int_field(self) -> int:
            return 2 * self.int_field

        @computed_field  # type: ignore[misc]
        @property
        def triple_int_field(self) -> int:
            return 3 * self.int_field

        int_field: int

    class ActiveDomainModel(ProvisioningDomainModel):
        @computed_field  # type: ignore[misc]
        @property
        def triple_int_field(self) -> int:
            # override the base property
            return 30 * self.int_field

    block = ActiveDomainModel(int_field=1)

    assert block.model_dump() == {"int_field": 1, "double_int_field": 2, "triple_int_field": 30}


def spy_get_subscription():
    # Spy on the _get_subscription method to observe the number of queries
    return mock.patch.object(
        SubscriptionModel,
        "_get_subscription",
        spec=SubscriptionModel._get_subscription,
        side_effect=SubscriptionModel._get_subscription,
    )


def test_nested_serializable_property():
    """Ensure that nested serializable property's are included in the serialized model."""

    class DerivedDomainModel(DomainModel):
        @computed_field  # type: ignore[misc]
        @property
        def double_int_field(self) -> int:
            # This property is serialized
            return 2 * self.int_field

        int_field: int

    class ParentDomainModel(DomainModel):
        derived: DerivedDomainModel

    model = ParentDomainModel(derived=DerivedDomainModel(int_field=13))

    assert model.model_dump() == {"derived": {"int_field": 13, "double_int_field": 26}}


def test_serializable_property_with_cache_subscription_models(
    generic_subscription_1, generic_subscription_2, generic_product_type_1, generic_product_type_2, test_product_model
):
    """Ensure that serializable properties can retrieve other SubscriptionModels without duplicate queries.

    While this works and will continue to work, is not recommended practice because reconstructing the entire
    subscription only to retrieve a few attributes is "overkill" and impacts peformance.
    Issue #899 has been created to track the design of a more efficient implementation.
    """
    _, GenericProductOne = generic_product_type_1
    _, GenericProductTwo = generic_product_type_2

    # One typical usecase -at SURF- is to define a computed 'title' for a product block based on details from
    # the owner subscription of a 'foreign' product block.
    #
    # For example:
    #
    # class PeerBlockProvisioning(PeerBlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
    #     port: IpPeerPortBlock
    #
    #     @computed_field
    #     @property
    #     def title(self) -> str:
    #         sap = self.port.sap
    #         subscription = SubscriptionModel.from_subscription(sap.owner_subscription_id)
    #         return f"IP Peering on {subscription.description} vlan {sap.vlanrange}"
    #
    # Below we test a simplified but technically identical usecase.

    # given

    class TestProduct(SubscriptionModel):
        @computed_field  # type: ignore
        @property
        def title(self) -> str:
            subscription1 = GenericProductOne.from_subscription(generic_subscription_1)
            subscription2 = GenericProductTwo.from_subscription(generic_subscription_2)
            return f"{subscription1.description} {subscription1.pb_2.rt_2} - {subscription2.description} {subscription2.pb_3.rt_2}"

        @computed_field  # type: ignore
        @property
        def foobar(self) -> str:
            subscription1 = GenericProductOne.from_subscription(generic_subscription_1)
            return f"{subscription1.description} {subscription1.pb_2.rt_2}"

    model = TestProduct(int_field=123, product=test_product_model, customer_id="")

    # when

    with spy_get_subscription() as mock_get_subscription_no_cache:
        actual_result_no_cache = model.model_dump(include=["title", "foobar"])

    with (
        spy_get_subscription() as mock_get_subscription_with_cache,
        cache_subscription_models(),  # <- cache enabled this time
    ):
        actual_result_with_cache = model.model_dump(include=["title", "foobar"])

    # then

    expected_result = {
        "title": "Generic Subscription One 42 - Generic Subscription Two 42",
        "foobar": "Generic Subscription One 42",
    }
    assert actual_result_no_cache == expected_result
    assert actual_result_with_cache == expected_result

    actual_calls_no_cache = sorted(mock_get_subscription_no_cache.mock_calls)
    expected_calls_no_cache = sorted(
        [
            mock.call(generic_subscription_1),
            mock.call(generic_subscription_1),
            mock.call(generic_subscription_2),
        ]
    )
    assert actual_calls_no_cache == expected_calls_no_cache

    actual_calls_with_cache = sorted(mock_get_subscription_with_cache.mock_calls)
    expected_calls_with_cache = sorted(
        [
            mock.call(generic_subscription_1),  # Queried once instead of twice
            mock.call(generic_subscription_2),
        ]
    )
    assert actual_calls_with_cache == expected_calls_with_cache



# Copyright 2019-2025 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Iterable

from sqlalchemy import Select, select

from orchestrator.db import (
    SubscriptionTable,
    WorkflowTable,
    db,
)
from orchestrator.schemas import StepSchema, WorkflowSchema
from orchestrator.services.subscriptions import TARGET_DEFAULT_USABLE_MAP, WF_USABLE_MAP
from orchestrator.targets import Target
from orchestrator.workflows import get_workflow


def _get_steps(workflow: WorkflowTable) -> list[StepSchema]:
    if registered_workflow := get_workflow(workflow.name):
        return [StepSchema(name=step.name) for step in registered_workflow.steps]
    raise AssertionError(f"Workflow {workflow.name} should be registered")


def _to_workflow_schema(workflow: WorkflowTable, include_steps: bool = False) -> WorkflowSchema:
    extra_kwargs = {"steps": _get_steps(workflow)} if include_steps else {}

    return WorkflowSchema(
        workflow_id=workflow.workflow_id,
        name=workflow.name,
        target=workflow.target,
        is_task=workflow.is_task,
        description=workflow.description,
        created_at=workflow.created_at,
        **extra_kwargs,
    )


def get_workflows(
    filters: dict | None = None, include_steps: bool = False, include_deleted: bool = False
) -> Iterable[WorkflowSchema]:
    def _add_filter(stmt: Select) -> Select:
        for k, v in (filters or {}).items():
            stmt = stmt.where(WorkflowTable.__dict__[k] == v)
        return stmt

    stmt = select(WorkflowTable) if include_deleted else WorkflowTable.select()
    workflows = db.session.scalars(_add_filter(stmt)).all()

    return [_to_workflow_schema(wf, include_steps=include_steps) for wf in workflows]


def get_workflow_by_name(workflow_name: str) -> WorkflowTable | None:
    return db.session.scalar(select(WorkflowTable).where(WorkflowTable.name == workflow_name))


def get_validation_product_workflows_for_subscription(
    subscription: SubscriptionTable,
) -> list:
    return [workflow.name for workflow in subscription.product.workflows if workflow.target == Target.VALIDATE]


def start_validation_workflow_for_workflows(
    subscription: SubscriptionTable,
    workflows: list,
    product_type_filter: str | None = None,
) -> list:
    """Start validation workflows for a subscription."""
    result = []

    for workflow_name in workflows:
        target_system = TARGET_DEFAULT_USABLE_MAP[Target.SYSTEM]
        system_usable_when = WF_USABLE_MAP.get(workflow_name, target_system)
        target_validate = TARGET_DEFAULT_USABLE_MAP[Target.VALIDATE]
        validate_usable_when = WF_USABLE_MAP.get(workflow_name, target_validate)

        usable_when = system_usable_when + validate_usable_when
        if subscription.status in usable_when and (
            product_type_filter is None or subscription.product.product_type == product_type_filter
        ):
            json = [{"subscription_id": str(subscription.subscription_id)}]

            # against circular import
            from orchestrator.services.processes import get_execution_context

            validate_func = get_execution_context()["validate"]
            validate_func(workflow_name, json=json)

            result.append(
                {
                    "workflow_name": workflow_name,
                    "subscription_id": subscription.subscription_id,
                    "product_type": subscription.product.product_type,
                }
            )

    return result



from http import HTTPStatus

import pytest

EMPTY_METRICS = """# HELP wfo_subscriptions_count Number of subscriptions per product, lifecycle state, customer, and in sync state.
# TYPE wfo_subscriptions_count gauge
# HELP wfo_process_count Number of processes per status, creator, task, product, workflow, customer, and target.
# TYPE wfo_process_count gauge
# HELP wfo_process_seconds_total_count Total time spent on processes in seconds.
# TYPE wfo_process_seconds_total_count gauge
# HELP wfo_engine_status Current workflow engine status.
# TYPE wfo_engine_status gauge
wfo_engine_status{wfo_engine_status="PAUSED"} 0.0
wfo_engine_status{wfo_engine_status="PAUSING"} 0.0
wfo_engine_status{wfo_engine_status="RUNNING"} 1.0
# HELP wfo_active_process_count Number of currently running processes in the workflow engine.
# TYPE wfo_active_process_count gauge
wfo_active_process_count 0.0
"""


@pytest.fixture(autouse=True, scope="function")
def assume_empty_metrics_at_start(test_client) -> None:
    """Assert that at the start of every unit test in this module, the metrics endpoint only contains empty data."""
    response = test_client.get("/api/metrics")
    assert HTTPStatus.OK == response.status_code
    assert response.text == EMPTY_METRICS



"""Add example1 product.

Revision ID: ea9e6c9de75c
Revises: 85be1c80731c
Create Date: 2024-02-20 21:01:56.190106

"""

from uuid import uuid4

from alembic import op
from orchestrator.migrations.helpers import create, create_workflow, delete, delete_workflow, ensure_default_workflows
from orchestrator.targets import Target

# revision identifiers, used by Alembic.
revision = "ea9e6c9de75c"
down_revision = "85be1c80731c"
branch_labels = None
depends_on = None

new_products = {
    "products": {
        "example1 1": {
            "product_id": uuid4(),
            "product_type": "Example1",
            "description": "Product example 1",
            "tag": "EXAMPLE1",
            "status": "active",
            "root_product_block": "Example1",
            "fixed_inputs": {
                "fixed_input_1": "1",
            },
        },
        "example1 10": {
            "product_id": uuid4(),
            "product_type": "Example1",
            "description": "Product example 1",
            "tag": "EXAMPLE1",
            "status": "active",
            "root_product_block": "Example1",
            "fixed_inputs": {
                "fixed_input_1": "10",
            },
        },
        "example1 100": {
            "product_id": uuid4(),
            "product_type": "Example1",
            "description": "Product example 1",
            "tag": "EXAMPLE1",
            "status": "active",
            "root_product_block": "Example1",
            "fixed_inputs": {
                "fixed_input_1": "100",
            },
        },
        "example1 1000": {
            "product_id": uuid4(),
            "product_type": "Example1",
            "description": "Product example 1",
            "tag": "EXAMPLE1",
            "status": "active",
            "root_product_block": "Example1",
            "fixed_inputs": {
                "fixed_input_1": "1000",
            },
        },
    },
    "product_blocks": {
        "Example1": {
            "product_block_id": uuid4(),
            "description": "Example 1 root product block",
            "tag": "EXAMPLE1",
            "status": "active",
            "resources": {
                "example_str_enum_1": "Example 1 str enum",
                "unmodifiable_str": "Unmodifiable resource type",
                "modifiable_boolean": "Modifiable resource type",
                "annotated_int": "Annotated integer witch min and max",
                "imported_type": "use imported type",
                "always_optional_str": "Not required in any lifecycle state",
            },
            "depends_on_block_relations": [
                "Example2",
            ],
        },
    },
    "workflows": {},
}

new_workflows = [
    {
        "name": "create_example1",
        "target": Target.CREATE,
        "is_task": False,
        "description": "Create example1",
        "product_type": "Example1",
    },
    {
        "name": "modify_example1",
        "target": Target.MODIFY,
        "is_task": False,
        "description": "Modify example1",
        "product_type": "Example1",
    },
    {
        "name": "terminate_example1",
        "target": Target.TERMINATE,
        "is_task": False,
        "description": "Terminate example1",
        "product_type": "Example1",
    },
    {
        "name": "validate_example1",
        "target": Target.VALIDATE,
        "is_task": True,
        "description": "Validate example1",
        "product_type": "Example1",
    },
]


def upgrade() -> None:
    conn = op.get_bind()
    create(conn, new_products)
    for workflow in new_workflows:
        create_workflow(conn, workflow)
    ensure_default_workflows(conn)


def downgrade() -> None:
    conn = op.get_bind()
    for workflow in new_workflows:
        delete_workflow(conn, workflow["name"])

    delete(conn, new_products)



from typing import Annotated

import pytest
import strawberry

from orchestrator.db import ProductBlockTable, db
from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SubscriptionLifecycle
from test.unit_tests.helpers import safe_delete_product_block_id


class ProductBlockListNestedForTestInactive(ProductBlockModel, product_block_name="ProductBlockListNestedForTest"):
    sub_block_list: list["ProductBlockListNestedForTestInactive"]
    int_field: int | None = None


class ProductBlockListNestedForTestProvisioning(
    ProductBlockListNestedForTestInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]
):
    sub_block_list: list["ProductBlockListNestedForTestProvisioning"]  # type: ignore
    int_field: int


class ProductBlockListNestedForTest(
    ProductBlockListNestedForTestProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]
):
    sub_block_list: list["ProductBlockListNestedForTest"]  # type: ignore
    int_field: int


ProductBlockListNestedForTestType = Annotated[
    "ProductBlockListNestedForTestInactiveGraphql", strawberry.lazy(".product_block_list_nested")
]


@strawberry.experimental.pydantic.type(model=ProductBlockListNestedForTestInactive)
class ProductBlockListNestedForTestInactiveGraphql:
    sub_block_list: list[ProductBlockListNestedForTestType]
    int_field: int


@pytest.fixture
def test_product_block_list_nested(test_product_block_list_nested_db_in_use_by_block):
    # Classes defined at module level, otherwise they remain in local namespace and
    # `get_type_hints()` can't evaluate the ForwardRefs
    yield (
        ProductBlockListNestedForTestInactive,
        ProductBlockListNestedForTestProvisioning,
        ProductBlockListNestedForTest,
    )

    safe_delete_product_block_id(ProductBlockListNestedForTestInactive)
    safe_delete_product_block_id(ProductBlockListNestedForTestProvisioning)
    safe_delete_product_block_id(ProductBlockListNestedForTest)


@pytest.fixture
def test_product_block_list_nested_db_in_use_by_block(resource_type_list, resource_type_int, resource_type_str):
    in_use_by_block = ProductBlockTable(
        name="ProductBlockListNestedForTest", description="Test Block Parent", tag="TEST", status="active"
    )
    in_use_by_block.resource_types = [resource_type_int]

    db.session.add(in_use_by_block)
    db.session.commit()

    return in_use_by_block



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from products.test_product import TestProduct

from orchestrator import OrchestratorCore
from orchestrator.cli.main import app as core_cli
from orchestrator.domain import SUBSCRIPTION_MODEL_REGISTRY
from orchestrator.services.subscriptions import WF_USABLE_MAP
from orchestrator.settings import AppSettings
from orchestrator.workflows import LazyWorkflowInstance

app = OrchestratorCore(base_settings=AppSettings())

SUBSCRIPTION_MODEL_REGISTRY.update({"TestProduct": TestProduct})

LazyWorkflowInstance("workflows.create_test_product", "create_test_product")
# LazyWorkflowInstance("workflows.terminate_test_product", "terminate_test_product")
# LazyWorkflowInstance("workflows.validate_test_product", "validate_test_product")

WF_USABLE_MAP.update(
    {
        # "validate_test_product": ["initial", "active", "provisioning"],
        "modify_note": ["initial", "active", "provisioning", "migrating", "terminated"],
    }
)

if __name__ == "__main__":
    core_cli()



from strawberry.types.nodes import FragmentSpread, InlineFragment, SelectedField

_LARGE_QUERY = """query GetWorkflows {
  workflows {
    page {
      ... on Workflow {
        aliasedName: name
      }
      ... MyFragment
      name
      products {
        tag
        productBlocks {
          name
          resourceTypes {
            description
            aliasedProductBlocks: productBlocks {
              name
            }
          }
          inUseBy {
            name
          }
        }
       }
    }
  }
  workflows {
    page {
      createdAt
    }
  }
}


fragment MyFragment on Workflow {
  aliasedTarget: target
}"""

# Generated by running the above _LARGE_QUERY and logging `info.selected_fields` from the resolver
# Note: this query is a bit silly, but it hits all the edgecases I can think of

LARGE_QUERY_SELECTED_FIELDS = [
    SelectedField(
        name="workflows",
        directives={},
        arguments={},
        selections=[
            SelectedField(
                name="page",
                directives={},
                arguments={},
                selections=[
                    InlineFragment(
                        type_condition="Workflow",
                        selections=[
                            SelectedField(name="name", directives={}, arguments={}, selections=[], alias="aliasedName")
                        ],
                        directives={},
                    ),
                    FragmentSpread(
                        name="MyFragment",
                        type_condition="Workflow",
                        directives={},
                        selections=[
                            SelectedField(
                                name="target", directives={}, arguments={}, selections=[], alias="aliasedTarget"
                            )
                        ],
                    ),
                    SelectedField(name="name", directives={}, arguments={}, selections=[], alias=None),
                    SelectedField(
                        name="products",
                        directives={},
                        arguments={},
                        selections=[
                            SelectedField(name="tag", directives={}, arguments={}, selections=[], alias=None),
                            SelectedField(
                                name="productBlocks",
                                directives={},
                                arguments={},
                                selections=[
                                    SelectedField(name="name", directives={}, arguments={}, selections=[], alias=None),
                                    SelectedField(
                                        name="resourceTypes",
                                        directives={},
                                        arguments={},
                                        selections=[
                                            SelectedField(
                                                name="description",
                                                directives={},
                                                arguments={},
                                                selections=[],
                                                alias=None,
                                            ),
                                            SelectedField(
                                                name="productBlocks",
                                                directives={},
                                                arguments={},
                                                selections=[
                                                    SelectedField(
                                                        name="name",
                                                        directives={},
                                                        arguments={},
                                                        selections=[],
                                                        alias=None,
                                                    )
                                                ],
                                                alias="aliasedProductBlocks",
                                            ),
                                        ],
                                        alias=None,
                                    ),
                                    SelectedField(
                                        name="inUseBy",
                                        directives={},
                                        arguments={},
                                        selections=[
                                            SelectedField(
                                                name="name", directives={}, arguments={}, selections=[], alias=None
                                            )
                                        ],
                                        alias=None,
                                    ),
                                ],
                                alias=None,
                            ),
                        ],
                        alias=None,
                    ),
                ],
                alias=None,
            )
        ],
        alias=None,
    ),
    SelectedField(
        name="workflows",
        directives={},
        arguments={},
        selections=[
            SelectedField(
                name="page",
                directives={},
                arguments={},
                selections=[SelectedField(name="createdAt", directives={}, arguments={}, selections=[], alias=None)],
                alias=None,
            )
        ],
        alias=None,
    ),
]


SMALL_QUERY_SELECTED_FIELDS = [
    LARGE_QUERY_SELECTED_FIELDS[1],
]



import structlog
from sqlalchemy import BinaryExpression

from orchestrator.db import ProductBlockTable, ProductTable, ResourceTypeTable
from orchestrator.db.filters import create_memoized_field_list, generic_filter_from_clauses
from orchestrator.db.filters.search_filters import default_inferred_column_clauses, node_to_str_val
from orchestrator.utils.search_query import Node

logger = structlog.get_logger(__name__)


def products_clause(node: Node) -> BinaryExpression:
    return ProductBlockTable.products.any(ProductTable.name.ilike(node_to_str_val(node)))


def resource_types_clause(node: Node) -> BinaryExpression:
    return ProductBlockTable.resource_types.any(ResourceTypeTable.resource_type.ilike(node_to_str_val(node)))


PRODUCT_BLOCK_TABLE_COLUMN_CLAUSES = default_inferred_column_clauses(ProductBlockTable) | {
    "product": products_clause,
    "resource_type": resource_types_clause,
}

product_block_filter_fields = create_memoized_field_list(PRODUCT_BLOCK_TABLE_COLUMN_CLAUSES)
filter_product_blocks = generic_filter_from_clauses(PRODUCT_BLOCK_TABLE_COLUMN_CLAUSES)



import sys
from functools import partial

import pytest
from typer.testing import CliRunner

from orchestrator.cli.database import app as db_app
from test.unit_tests.cli.helpers import create_main


@pytest.fixture(scope="module")
def monkey_module():
    with pytest.MonkeyPatch.context() as mp:
        yield mp


@pytest.fixture(scope="module")
def tmp_generate_path(tmp_path_factory):
    yield tmp_path_factory.mktemp("generate")


@pytest.fixture(scope="module")
def cli_invoke(tmp_generate_path, monkey_module):
    monkey_module.chdir(tmp_generate_path)
    sys.path.append(str(tmp_generate_path))
    create_main()

    runner = CliRunner()
    # Don't catch exceptions because this will cost you grey hair.
    invoke = partial(runner.invoke, catch_exceptions=False)
    invoke(db_app, ["init"])

    yield invoke



# Copyright 2019-2020 SURF, GÉANT.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Module that implements subscription related API endpoints."""
from http import HTTPStatus
from typing import Any
from uuid import UUID

import structlog
from fastapi import Depends
from fastapi.routing import APIRouter
from sqlalchemy import delete, select
from sqlalchemy.orm import contains_eager, defer, joinedload
from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import Response

from oauth2_lib.fastapi import OIDCUserModel
from orchestrator.api.error_handling import raise_status
from orchestrator.api.helpers import add_response_range, add_subscription_search_query_filter
from orchestrator.db import (
    ProcessStepTable,
    ProcessSubscriptionTable,
    ProcessTable,
    ProductTable,
    SubscriptionMetadataTable,
    SubscriptionTable,
    db,
)
from orchestrator.schemas import SubscriptionWorkflowListsSchema
from orchestrator.schemas.subscription import SubscriptionDomainModelSchema, SubscriptionWithMetadata
from orchestrator.security import authenticate
from orchestrator.services.subscriptions import (
    format_extended_domain_model,
    format_special_types,
    get_subscription,
    subscription_workflows,
)
from orchestrator.settings import app_settings
from orchestrator.types import SubscriptionLifecycle
from orchestrator.utils.deprecation_logger import deprecated_endpoint
from orchestrator.utils.get_subscription_dict import get_subscription_dict
from orchestrator.websocket import sync_invalidate_subscription_cache

router = APIRouter()

logger = structlog.get_logger(__name__)


def _delete_subscription_tree(subscription: SubscriptionTable) -> None:
    db.session.delete(subscription)
    db.session.commit()


def _delete_process_subscriptions(process_subscriptions: list[ProcessSubscriptionTable]) -> None:
    for process_subscription in process_subscriptions:
        process_id = str(process_subscription.process_id)
        subscription_id = str(process_subscription.subscription_id)
        db.session.execute(delete(ProcessSubscriptionTable).filter(ProcessSubscriptionTable.process_id == process_id))
        db.session.execute(delete(ProcessStepTable).filter(ProcessStepTable.process_id == process_id))
        db.session.execute(delete(ProcessTable).filter(ProcessTable.process_id == process_id))
        subscription = db.session.scalars(
            select(SubscriptionTable).filter(SubscriptionTable.subscription_id == subscription_id)
        ).first()
        if subscription:
            _delete_subscription_tree(subscription)


def _filter_statuses(filter_statuses: str | None = None) -> list[str]:
    """Check valid filter statuses.

    Args:
        filter_statuses: the filters.

    Returns:
        list of filters

    """
    if not filter_statuses:
        return []

    logger.debug("Filters to query subscription on.", filter_statuses=filter_statuses)
    statuses = filter_statuses.split(",")
    for status in statuses:
        if status not in SubscriptionLifecycle.values():
            raise HTTPException(
                status_code=HTTPStatus.UNPROCESSABLE_ENTITY,
                detail=f"Status {status}, is not a valid `SubscriptionLifecycle`",
            )
    return statuses


@router.get(
    "/domain-model/{subscription_id}",
    response_model=SubscriptionDomainModelSchema | None,
)
async def subscription_details_by_id_with_domain_model(
    request: Request, subscription_id: UUID, response: Response, filter_owner_relations: bool = True
) -> dict[str, Any] | None:
    def _build_response(model: dict, etag: str) -> dict[str, Any] | None:
        if etag == request.headers.get("If-None-Match"):
            response.status_code = HTTPStatus.NOT_MODIFIED
            return None
        response.headers["ETag"] = etag
        filtered = format_extended_domain_model(model, filter_owner_relations=filter_owner_relations)
        return format_special_types(filtered)

    try:
        subscription, etag = await get_subscription_dict(subscription_id)
        return _build_response(subscription, etag)
    except ValueError as e:
        if str(e) == f"Subscription with id: {subscription_id}, does not exist":
            raise_status(HTTPStatus.NOT_FOUND, f"Subscription with id: {subscription_id}, not found")
        else:
            raise_status(HTTPStatus.INTERNAL_SERVER_ERROR, str(e))


@router.get(
    "/search",
    response_model=list[SubscriptionWithMetadata],
)
def subscriptions_search(
    response: Response, query: str, range: str | None = None, sort: str | None = None
) -> list[dict]:
    """Get subscriptions filtered based on a search query string.

    Args:
        response: Fastapi Response object
        query: The search query
        range: Range
        sort: Sort

    Returns:
        List of subscriptions

    """
    range_ = list(map(int, range.split(","))) if range else None
    sort_ = sort.split(",") if sort else None
    logger.info("subscriptions_search() called", range=range_, sort=sort_)
    stmt = select(SubscriptionTable, SubscriptionMetadataTable.metadata_).join_from(
        SubscriptionTable, SubscriptionMetadataTable, isouter=True
    )

    stmt = stmt.join(SubscriptionTable.product).options(
        contains_eager(SubscriptionTable.product), defer(SubscriptionTable.product_id)
    )
    stmt = add_subscription_search_query_filter(stmt, query)
    stmt = add_response_range(stmt, range_, response, unit="subscriptions")
    sequence = db.session.execute(stmt).all()
    return [{**s.__dict__, "metadata": md} for s, md in sequence]


@router.get(
    "/workflows/{subscription_id}",
    response_model=SubscriptionWorkflowListsSchema,
    response_model_by_alias=True,
    response_model_exclude_none=True,
    deprecated=True,
    description="This endpoint is deprecated and will be removed in a future release. Please use the GraphQL query",
    dependencies=[Depends(deprecated_endpoint)],
)
def subscription_workflows_by_id(subscription_id: UUID) -> dict[str, list[dict[str, list[Any] | str]]]:
    subscription = db.session.get(
        SubscriptionTable,
        subscription_id,
        options=[
            joinedload(SubscriptionTable.product),
            joinedload(SubscriptionTable.product).joinedload(ProductTable.workflows),
        ],
    )
    if not subscription:
        raise_status(HTTPStatus.NOT_FOUND)

    return subscription_workflows(subscription)


@router.put("/{subscription_id}/set_in_sync", response_model=None, status_code=HTTPStatus.OK)
def subscription_set_in_sync(subscription_id: UUID, current_user: OIDCUserModel | None = Depends(authenticate)) -> None:
    def failed_processes() -> list[str]:
        if app_settings.DISABLE_INSYNC_CHECK:
            return []
        stmt = (
            select(ProcessSubscriptionTable)
            .join(ProcessTable)
            .filter(ProcessSubscriptionTable.subscription_id == subscription_id)
            .filter(~ProcessTable.is_task)
            .filter(ProcessTable.last_status != "completed")
            .filter(ProcessTable.last_status != "aborted")
        )
        _failed_processes = db.session.scalars(stmt)
        return [str(p.process_id) for p in _failed_processes]

    try:
        subscription = get_subscription(subscription_id, for_update=True)
        if not subscription.insync:
            logger.info(
                "Subscription not in sync, trying to change..", subscription_id=subscription_id, user=current_user
            )
            failed_processes_list = failed_processes()
            if not failed_processes_list:
                subscription.insync = True
                db.session.commit()
                logger.info("Subscription set in sync", user=current_user)
                sync_invalidate_subscription_cache(subscription.subscription_id)
            else:
                raise_status(
                    HTTPStatus.UNPROCESSABLE_ENTITY,
                    f"Subscription {subscription_id} has still failed processes with id's: {failed_processes_list}",
                )
        else:
            logger.info("Subscription already in sync")
    except ValueError as e:
        raise_status(HTTPStatus.NOT_FOUND, str(e))



import json
from http import HTTPStatus


def get_version_query() -> bytes:
    query = """
        query VersionQuery {
        version {
            applicationVersions
        }
        }
    """
    return json.dumps({"operationName": "VersionQuery", "query": query}).encode("utf-8")


def test_version_query(test_client):
    data = get_version_query()
    response = test_client.post("/api/graphql", content=data, headers={"Content-Type": "application/json"})
    assert response.status_code == HTTPStatus.OK
    result = response.json()
    assert "errors" not in result



from orchestrator.domain.base import ProductBlockModel
from orchestrator.types import SubscriptionLifecycle
from pydantic import computed_field

from products.product_blocks.example4sub import Example4SubBlock, Example4SubBlockInactive, Example4SubBlockProvisioning


class Example4BlockInactive(ProductBlockModel, product_block_name="Example4"):
    num_val: int | None = None
    sub_block: Example4SubBlockInactive | None = None


class Example4BlockProvisioning(Example4BlockInactive, lifecycle=[SubscriptionLifecycle.PROVISIONING]):
    num_val: int | None = None
    sub_block: Example4SubBlockProvisioning

    @computed_field
    @property
    def title(self) -> str:
        # TODO: format correct title string
        return f"{self.name}"


class Example4Block(Example4BlockProvisioning, lifecycle=[SubscriptionLifecycle.ACTIVE]):
    num_val: int | None = None
    sub_block: Example4SubBlock



import pytest

from test.unit_tests.workflows import assert_complete, extract_state, run_workflow


@pytest.mark.workflow
def test_happy_flow(responses, example4_subscription):
    # when

    result, _, _ = run_workflow("validate_example4", {"subscription_id": example4_subscription})

    # then

    assert_complete(result)
    state = extract_state(result)
    assert state["check_core_db"] is True



# Copyright 2019-2020 SURF.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
from typing import Any
from uuid import UUID

from pydantic import ConfigDict, Field

from orchestrator.schemas.base import OrchestratorBaseModel
from orchestrator.targets import Target


class WorkflowBaseSchema(OrchestratorBaseModel):
    name: str
    target: Target
    is_task: bool = False
    description: str | None = None
    created_at: datetime | None = None


class StepSchema(OrchestratorBaseModel):
    name: str


class WorkflowSchema(WorkflowBaseSchema):
    workflow_id: UUID
    created_at: datetime
    steps: list[StepSchema] | None = None

    model_config = ConfigDict(from_attributes=True)


class WorkflowListItemSchema(OrchestratorBaseModel):
    name: str
    description: str | None = None
    reason: str | None = None
    usable_when: list[Any] | None = None
    status: str | None = None
    action: str | None = None
    locked_relations: list[UUID] | None = None
    unterminated_parents: list[UUID] | None = None
    unterminated_in_use_by_subscriptions: list[UUID] | None = None


class SubscriptionWorkflowListsSchema(OrchestratorBaseModel):
    reason: str | None = None
    locked_relations: list[UUID] | None = None
    create: list[WorkflowListItemSchema]
    modify: list[WorkflowListItemSchema]
    terminate: list[WorkflowListItemSchema]
    system: list[WorkflowListItemSchema]
    validate_: list[WorkflowListItemSchema] = Field(default_factory=list, alias="validate")


class WorkflowPatchSchema(OrchestratorBaseModel):
    description: str | None = None
